SystemOrganization addCategory: #'Movies-Player'!
SystemOrganization addCategory: #'Movies-Kernel'!

BookMorph subclass: #MoviePlayerMorph
	instanceVariableNames: 'movieFileName movieFile frameSize frameDepth frameNumber frameCount playDirection msSinceStart msAtStart msAtLastSync frameAtLastSync msPerFrame frameBufferIfScaled soundTrackFileName scorePlayer soundTrackForm soundTrackMorph pianoRoll cueMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Player'!

!MoviePlayerMorph commentStamp: '<historical>' prior: 0!
MoviePlayerMorph plays images from a file using async io.  The file format is simple but non-standard (see below).

The heart of the play logic is in the step method.  Note that play is driven by a simulated time since start.  For a movie with a sound score, this is the millisecondsSinceStart of the score player, whereas a movie by itself gets this from the millisecondClock minus msAtStart.

Movie players are designed to be used in three ways
	1.  Select a movie in the file list, 'open as movie', and play it.
	2.  As in (1), but drop thumbnails for various frames into a
		MIDI piano roll to synchronize video with music.
	3.  Open a MoviePlayerMorph as a 'new morph', and choose
		'make a new movie' from the menu.

In (1) and (2), a shift-drag is used to 'tear off' a thumbnail reference morph to the currently visible frame of this clips.  The thumbnail can then be dropped in a MIDI score player to either syncronize that frame with what point in the music, or to cause that clip to being playing in the shared player of a SqueakMovie plex.

When making a new movie, an empty score and piano roll are linked to the movie player, as a SqueakMovie plex. You can then open another movie as a clip, and drop a thumbnail morph into the score to start that clip playing at that frame in the shared player.  If you pause while playing that clip, you can manually play the clip forward and backward in the current clip.  if you stop at a given frame, you can choose 'end clip here' from the shared player menu to shorten or lengthen the clip.

Clips can be moved by picking up the starting thumbnail (use halo black handle), and dropping them elsewhere.  If you try to place one clip in the middle of another, it will slide to the end.  If you position one clip close to the end of another, it will sidle up to make them contiguous.

If you wish a soundtrack to be included with a clip, make sure it has been opened in the source clip player before tearing off the starting thumbnail.

About the .movie file format...
The following code was used to convert 27 files into a movie.  They were named
	'BalloonDrop10fps003.bmp' through 'BalloonDrop10fps081.bmp'
incrementing by 003.  Each was known to be a 320x240 image in 32-bit BMP format.
Note the 27 in the 5th line is the number of frames, = (3 to: 81 by: 3) size.

	| ps zps f32 out ff |
	out _ FileStream newFileNamed: 'BalloonDrop10fps.movie'.
	out binary.
	ff _ Form extent: 320@240 depth: 16.
	#(22 320 240 16 27 100000) , (7 to: 32)
		do: [:i | out nextInt32Put: i].
		
	3 to: 81 by: 3 do:
		[:i | ps _ i printString. zps _ ps padded: #left to: 3 with: $0.
		f32 _ Form fromFileNamed:
			'BalloonDrop10fps' , zps , '.bmp'.
		f32 displayOn: ff at: 0@0.  "Convert down to 16 bits"
		ff display; writeOnMovie: out].
	out close.
!

!MoviePlayerMorph class methodsFor: 'fileIn/Out' stamp: 'sd 2/6/2002 21:36'!
fileReaderServicesForFile: fullName suffix: suffix

	^(suffix = 'movie') | (suffix = '*')
		ifTrue: [ self services]
		ifFalse: [#()]! !

!MoviePlayerMorph class methodsFor: 'class initialization' stamp: 'hg 8/3/2000 16:59'!
initialize

	FileList registerFileReader: self! !

!MoviePlayerMorph class methodsFor: 'fileIn/Out' stamp: 'hg 8/3/2000 17:01'!
openAsMovie: fullFileName
	"Open a MoviePlayerMorph on the given file (must be in .movie format)."
 
	(self new openFileNamed: fullFileName) openInWorld! !

!MoviePlayerMorph class methodsFor: 'fileIn/Out' stamp: 'sw 2/17/2002 01:23'!
serviceOpenAsMovie
	"Answer a service for opening a file as a movie"

	^ SimpleServiceEntry 
		provider: self 
		label: 'open as movie'
		selector: #openAsMovie:
		description: 'open file as movie'
		buttonLabel: 'open'! !

!MoviePlayerMorph class methodsFor: 'fileIn/Out' stamp: 'sw 9/7/2004 18:21'!
services
	"Formerly: answer a service for opening as a movie.  Nowadays... no services"

	^ #().
"
	^ Array with: self serviceOpenAsMovie"

	! !

!MoviePlayerMorph class methodsFor: 'initialize-release' stamp: 'SD 11/15/2001 22:22'!
unload

	FileList unregisterFileReader: self ! !

!MoviePlayerMorph methodsFor: 'menu' stamp: 'dgd 2/22/2003 13:21'!
addSoundTrack
	| fileName |
	fileName := Utilities chooseFileWithSuffixFromList: #('.aif' '.wav')
				withCaption: 'Choose a sound track file'.
	fileName isNil ifTrue: [^self].
	soundTrackFileName := fileName.
	self tryToShareScoreFor: soundTrackFileName.
	scorePlayer ifNil: 
			[('*aif' match: fileName) 
				ifTrue: [scorePlayer := SampledSound fromAIFFfileNamed: fileName].
			('*wav' match: fileName) 
				ifTrue: [scorePlayer := SampledSound fromWaveFileNamed: fileName]].
	soundTrackForm ifNotNil: 
			["Compute new soundTrack if we're showing it."

			self
				showHideSoundTrack;
				showHideSoundTrack]! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/26/2000 00:18'!
cueMorph
	^ cueMorph! !

!MoviePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:38'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 2! !

!MoviePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:29'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color veryLightGray! !

!MoviePlayerMorph methodsFor: 'copying' stamp: 'di 10/30/2000 08:10'!
duplicate
	| dup |
	playDirection ~= 0 ifTrue: [self stopPlay].
	dup _ super duplicate.
	dup scorePlayer: scorePlayer copy.  "Share sound track if any."
	^ dup duplicateMore startStepping! !

!MoviePlayerMorph methodsFor: 'initialization' stamp: 'di 10/30/2000 08:11'!
duplicateMore
	"Duplicate dies not replicate Forms, but MoviePlayers need this."

	frameBufferIfScaled _ frameBufferIfScaled deepCopy.
	currentPage image: currentPage image deepCopy.
! !

!MoviePlayerMorph methodsFor: 'menu' stamp: 'di 10/19/2000 17:07'!
endClipHere
	"Change set the termination time for this clip via the endMorph"

	cueMorph ifNil: [^ self].
	cueMorph setEndFrameNumber: frameNumber
! !

!MoviePlayerMorph methodsFor: 'geometry' stamp: 'di 10/29/2000 12:02'!
extent: newExtent

	| tlMargin brMargin pageExtent scale fullSizeImage |
	fullSizeImage _ frameBufferIfScaled ifNil: [currentPage image].
	frameCount ifNil: [^ self].  "Not yet open"
	tlMargin _ currentPage topLeft - self topLeft.
	brMargin _ self bottomRight - currentPage bottomRight.
	pageExtent _ newExtent - brMargin - tlMargin.
	scale _ pageExtent x asFloat / frameSize x min: pageExtent y asFloat / frameSize y.
	(scale _ scale max: 0.25) > 0.9 ifTrue: [scale _ 1.0].

	pageExtent _ (frameSize * scale) rounded.
	pageExtent = frameSize
		ifTrue: [currentPage image: fullSizeImage.
				frameBufferIfScaled _ nil]
		ifFalse: [currentPage image: (Form extent: pageExtent depth: frameDepth).
				frameBufferIfScaled _ fullSizeImage.
				(WarpBlt current toForm: currentPage image) sourceForm: fullSizeImage;
					combinationRule: 3;
					copyQuad: fullSizeImage boundingBox innerCorners
						toRect: currentPage image boundingBox].
	^ self layoutChanged
! !

!MoviePlayerMorph methodsFor: 'stepping' stamp: 'di 9/5/2000 23:09'!
fileByteCountPerFrame
	
	^ (frameBufferIfScaled ifNil: [currentPage image]) bits size * 4
! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/4/2000 17:19'!
fileName
	^ movieFileName! !

!MoviePlayerMorph methodsFor: 'stepping' stamp: 'di 5/27/1999 22:40'!
filePosForFrameNo: frameNo
	
	^ 128 + ((frameNo-1)*(4+self fileByteCountPerFrame)) + 4
! !

!MoviePlayerMorph methodsFor: 'menu commands' stamp: 'di 7/6/1998 19:32'!
firstPage
	playDirection = 0 ifFalse: [^ self]. "No-op during play"
	self goToPage: 1
! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/22/2000 10:54'!
frameCount

	^ frameCount! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/22/2000 16:01'!
frameNumber: n

	frameNumber _ n! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 6/1/1999 09:13'!
frameNumber: n msSinceStart: ms
	"Set the current frame number, and save the scorePlayer's simulated time for synchronization."

	frameAtLastSync _ n.
	msAtLastSync _ ms.
! !

!MoviePlayerMorph methodsFor: 'page controls' stamp: 'di 10/16/2000 13:20'!
fullControlSpecs
	^ #(	
			( ''		invokeBookMenu 'Invoke menu')
			( '<--'		firstPage		'Go to first page')
			( '<<'		playReverse		'Play backward')
			( '<-' 		previousPage	'Back one frame')
			( '| |' 		stopPlay		'Stop playback')
			( '->'		nextPage		'Forward one frame')
			( '>>'		playForward	'Play forward')
			( '-->'		lastPage			'Go to final page')
			( '<->'		scanBySlider	'Scan by slider'  'menu')
		"Note extra spec 'menu' causes mousedown activation -- see makePageControlsFrom:"
	)! !

!MoviePlayerMorph methodsFor: 'navigation' stamp: 'md 12/12/2003 16:21'!
goToPage: i
	currentPage ifNil: [self makeMyPage].
	frameNumber _ i.
	playDirection _ 0.
	self startRunning; step.  "will stop after first step"
	soundTrackMorph ifNotNilDo: [:m | m image fillWhite].
	self stepSoundTrack.
! !

!MoviePlayerMorph methodsFor: 'player control' stamp: 'di 10/16/2000 11:02'!
goToRelativePosition: newPos

	movieFile ifNotNil: [^ self].
	self goToPage: ((newPos*frameCount) asInteger min: frameCount max: 1).
! !

!MoviePlayerMorph methodsFor: 'event handling' stamp: 'di 10/16/2000 13:27'!
handlesMouseDown: evt
	"We use shift drag to 'tear off' a thumbnail"

	evt shiftPressed ifTrue: [^ true].
	^ super handlesMouseDown: evt! !

!MoviePlayerMorph methodsFor: 'menu commands' stamp: 'di 7/6/1998 14:05'!
insertPage
	^ self makeMyPage! !

!MoviePlayerMorph methodsFor: 'menu' stamp: 'dgd 10/8/2003 19:32'!
invokeBookMenu
	"Invoke the book's control panel menu."
	| aMenu |
	aMenu _ MVCMenuMorph new defaultTarget: self.
	aMenu add:	'make a new movie' translated action: #makeAMovie.
	aMenu add:	'open movie file' translated action: #openMovieFile.
	aMenu add:	'add sound track' translated action: #addSoundTrack.
	aMenu addLine.
	scorePlayer ifNotNil:
		[soundTrackForm isNil
			ifTrue: [aMenu add:	'show sound track' translated action: #showHideSoundTrack]
			ifFalse: [aMenu add:	'hide sound track' translated action: #showHideSoundTrack]].
	aMenu add:	'make thumbnail' translated action: #thumbnailForThisPage.
	cueMorph ifNotNil:
		["Should check if piano roll and score already have a start event
		prior to this time."
		aMenu add:	'end clip here' translated action: #endClipHere].

	aMenu popUpEvent: self world activeHand lastEvent in: self world
! !

!MoviePlayerMorph methodsFor: 'navigation' stamp: 'di 7/6/1998 19:29'!
lastPage
	playDirection = 0 ifFalse: [^ self]. "No-op during play"
	self goToPage: frameCount
! !

!MoviePlayerMorph methodsFor: 'menu' stamp: 'di 10/24/2000 17:02'!
makeAMovie
	| scoreController score |
	frameSize _ 640@480.  frameDepth _ 16.  self makeMyPage; changed.

	(score _ MIDIScore new initialize) "addAmbientEvent: (AmbientEvent new time: 200*60)".
	scoreController _ ScorePlayerMorph new
			onScorePlayer: (ScorePlayer onScore: score) title: 'sMovie'.
	pianoRoll _ PianoRollScoreMorph new on: scoreController scorePlayer.
	self pianoRoll: pianoRoll.  "back link"
	pianoRoll enableDragNDrop;
		useRoundedCorners;
		movieClipPlayer: self;
		borderWidth: 2;
		extent: self width @ 120;
		align: pianoRoll topLeft with: self bottomLeft - (0@2);
		openInWorld.
	scoreController extent: self width @ scoreController height;
		align: scoreController topLeft with: pianoRoll bottomLeft - (0@2);
		openInWorld.

! !

!MoviePlayerMorph methodsFor: 'private' stamp: 'di 10/24/2000 16:59'!
makeMyPage

	currentPage ifNotNil:
		[(currentPage isMemberOf: ImageMorph)
			ifTrue: ["currentPage is already an ImageMorph."
					(currentPage image extent = frameSize
						and: [currentPage image depth = frameDepth])
						ifTrue: [^ self  "page is already properly dimensioned."].
					^ currentPage image: (Form extent: frameSize depth: frameDepth)]
			ifFalse: [currentPage releaseCachedState; delete]].
	currentPage _ ImageMorph new image: (Form extent: frameSize depth: frameDepth).
	currentPage lock.
	pages _ OrderedCollection with: currentPage.
	self addMorphFront: currentPage! !

!MoviePlayerMorph methodsFor: 'private' stamp: 'di 10/25/2000 23:01'!
makeThumbnailInHand: aHand

	scorePlayer ifNotNil:
		["Position the soundTrack for this frameNumber"
		scorePlayer reset; playSilentlyUntil: frameNumber - 1 * msPerFrame / 1000.0].

	aHand attachMorph:
		(MovieClipStartMorph new
			movieFileName: movieFileName
			soundTrackFileName: soundTrackFileName
			image: currentPage image
			player: self
			frameNumber: frameNumber)
! !

!MoviePlayerMorph methodsFor: 'event handling' stamp: 'di 10/16/2000 13:32'!
mouseDown: evt
	"We use shift drag to 'tear off' a thumbnail"

	evt shiftPressed ifTrue: [^ self makeThumbnailInHand: evt hand].
	^ super mouseDown: evt
		! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/19/2000 16:31'!
msPerFrame
	^ msPerFrame! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 6/1/1999 08:55'!
msPerFrame: n
	"Set a new frame rate, base on, eg, score synchronization info."

	msPerFrame _ n.
! !

!MoviePlayerMorph methodsFor: 'navigation' stamp: 'di 7/6/1998 19:29'!
nextPage
	playDirection = 0 ifFalse: [^ self]. "No-op during play"
	self goToPage: (frameNumber _ frameNumber + 1 min: frameCount).
! !

!MoviePlayerMorph methodsFor: 'initialization' stamp: 'di 10/8/2000 21:43'!
openFileNamed: fName 
	self pvtOpenFileNamed: fName.
	self goToPage: 1! !

!MoviePlayerMorph methodsFor: 'initialization' stamp: 'di 10/27/2000 15:22'!
openFileNamed: fName withScorePlayer: playerReady andPlayFrom: frameNo
	"Note: The plan is that the score player (a SampledSound) is already spaced
	forward to this frame number so it does not need to be reset as would normally
	happen in startRunning."

	self pvtOpenFileNamed: fName.
	scorePlayer _ playerReady.
	frameNumber _ frameNo.
	frameAtLastSync _ frameNo.
	msAtLastSync _ frameAtLastSync * msPerFrame.
	self playForward.! !

!MoviePlayerMorph methodsFor: 'menu' stamp: 'di 9/7/2000 09:44'!
openMovieFile
	| fileName |
	fileName _ Utilities chooseFileWithSuffixFromList: #('.movie')
					withCaption: 'Choose a movie file to open'.
	fileName ifNotNil:
		[self openFileNamed: fileName.
		self showMoreControls]! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/23/2000 01:45'!
pageFormForFrame: frameNo

	| f form oldFrame |
oldFrame _ frameNumber.
self goToPage: frameNo.
form _ currentPage image deepCopy.
self goToPage: oldFrame.
true ifTrue: [^ form].

	f _ FileStream readOnlyFileNamed: movieFileName.
	form _ Form extent: frameSize depth: frameDepth.

	"For some weird reason, the next line does not work..."
	f position: (self filePosForFrameNo: frameNo).
	"... but this line was found empirically to work instead."
	f position: (128 + ((frameNo-1)*(form bits size*4+4)) + 4).

	f nextInto: form bits.
	f close.
	^ form! !

!MoviePlayerMorph methodsFor: 'piano rolls' stamp: 'di 10/25/2000 22:04'!
pauseFrom: player

	playDirection _ 0.
	self step.! !

!MoviePlayerMorph methodsFor: 'movie clip player' stamp: 'di 10/19/2000 22:02'!
pianoRoll
	^ pianoRoll! !

!MoviePlayerMorph methodsFor: 'movie clip player' stamp: 'di 10/4/2000 16:38'!
pianoRoll: aPianoRollScoreMorph
	"Provides access also to the score and scorePlayer"

	pianoRoll _ aPianoRollScoreMorph
! !

!MoviePlayerMorph methodsFor: 'player control' stamp: 'di 8/17/1998 21:30'!
playForward
	(playDirection ~= 0 or: [frameNumber >= frameCount]) ifTrue:
		[^ self]. "No-op during play or at end"
	playDirection _ 1.
	self startRunning! !

!MoviePlayerMorph methodsFor: 'player control' stamp: 'di 8/17/1998 21:30'!
playReverse
	(playDirection ~= 0 or: [frameNumber <= 1]) ifTrue:
		[^ self]. "No-op during play or at end"
	playDirection _ -1.
	self startRunning! !

!MoviePlayerMorph methodsFor: 'geometry' stamp: 'dgd 2/22/2003 19:01'!
position: newPos 
	super position: newPos.
	(currentPage notNil and: [currentPage left odd]) 
		ifTrue: 
			["crude word alignment for depth = 16"

			super position: newPos + (1 @ 0)]! !

!MoviePlayerMorph methodsFor: 'navigation' stamp: 'di 7/6/1998 19:29'!
previousPage
	playDirection = 0 ifFalse: [^ self]. "No-op during play"
	self goToPage: (frameNumber _ frameNumber - 1 max: 1).
! !

!MoviePlayerMorph methodsFor: 'private' stamp: 'md 10/26/2003 13:07'!
pvtOpenFileNamed: fName
	"Private - open on the movie file iof the given name"

	| f w h d n m |
	self stopRunning.
	fName = movieFileName ifTrue: [^ self].  "No reopen necessary on same file"

	movieFileName _ fName.
	"Read movie file parameters from 128-byte header...
		(records follow as {N=int32, N words}*)"
	f _ (FileStream oldFileNamed: movieFileName) binary.
		f nextInt32.
		w _ f nextInt32.
		h _ f nextInt32.
		d _ f nextInt32.
		n _ f nextInt32.
		m _ f nextInt32.
		f close.
	pageSize _ frameSize _ w@h.
	frameDepth _ d.
	frameCount _ n.
	frameNumber _ 1.
	playDirection _ 0.
	msAtLastSync _ 0.
	msPerFrame _ m/1000.0.
	self makeMyPage.
	(SmalltalkImage current platformName = 'Mac OS') ifTrue:[
		(SmalltalkImage current extraVMMemory < self fileByteCountPerFrame) ifTrue:
			[^ self inform:
'Playing movies in Squeak requires that extra memory be allocated
for asynchronous file IO.  This particular movie requires a buffer of
' ,
(self fileByteCountPerFrame printString) , ' bytes, but you only have ' , (SmalltalkImage current extraVMMemory printString) , ' allocated.
You can evaluate ''SmalltalkImage current extraVMMemory'' to check your allocation,
and ''SmalltalkImage current extraVMMemory: 485000'' or the like to increase your allocation.
Note that raising your allocation in this way only marks your image as
needing this much, so you must then save, quit, and start over again
before you can run this movie.  Good luck.']].
! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/15/2000 19:15'!
relativePosition

	^ frameNumber asFloat / frameCount! !

!MoviePlayerMorph methodsFor: 'piano rolls' stamp: 'di 10/25/2000 22:06'!
resetFrom: player
	self pauseFrom: player.
	self firstPage! !

!MoviePlayerMorph methodsFor: 'piano rolls' stamp: 'di 10/22/2000 12:56'!
resumeFrom: player
	playDirection ~= 0 ifTrue: [^ self].  "Already running"
	playDirection _ 1.
	pianoRoll ifNil:
		["Sync movie to score player if not a clip player"
		scorePlayer _ player].
	self startRunning! !

!MoviePlayerMorph methodsFor: 'menu' stamp: 'di 10/16/2000 13:21'!
scanBySlider
	| scrollSlider handle |
	scrollSlider _ SimpleSliderMorph new extent: 150@10;
		color: color; sliderColor: Color gray;
		target: self; actionSelector: #goToRelativePosition:;
		adjustToValue: self relativePosition.
	(handle _ scrollSlider firstSubmorph) on: #mouseUp send: #delete to: scrollSlider.
	scrollSlider align: handle center with: self activeHand position.
	self world addMorph: scrollSlider.
	self activeHand targetOffset: (handle width // 2) @ 0.
	self activeHand newMouseFocus: handle! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/11/2000 12:13'!
scorePlayer

	^ scorePlayer! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/11/2000 12:20'!
scorePlayer: aScorePlayer

	scorePlayer _ aScorePlayer! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/19/2000 10:27'!
setCueMorph: aMorph
	cueMorph _ aMorph
! !

!MoviePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:07'!
setInitialState
	super setInitialState.
""
	self layoutInset: 3.
	pageSize _ frameSize _ 200 @ 200.
	frameDepth _ 8.
	self disableDragNDrop! !

!MoviePlayerMorph methodsFor: 'controls and layout' stamp: 'di 10/24/2000 17:22'!
showHideSoundTrack

	soundTrackForm ifNotNil:
		[soundTrackMorph delete.
		^ soundTrackForm _ soundTrackMorph _ nil].

	soundTrackForm _ scorePlayer volumeForm: 20 from: 1 to: scorePlayer samples size nSamplesPerPixel: 250.
	soundTrackMorph _ ImageMorph new image: (Form extent: 140 @ soundTrackForm height).
	soundTrackMorph addMorph:
		(Morph newBounds: (soundTrackMorph bounds topCenter extent: 1@soundTrackMorph height)
					color: Color red).
	self addMorph: soundTrackMorph after: currentPage.
	self layoutChanged.
	self stepSoundTrack.
! !

!MoviePlayerMorph methodsFor: 'page controls' stamp: 'di 10/24/2000 16:55'!
showPageControls
	self showPageControls: self fullControlSpecs.
! !

!MoviePlayerMorph methodsFor: 'page controls' stamp: 'ar 11/9/2000 20:42'!
showPageControls: controlSpecs 
	| pageControls |
	self hidePageControls.

	pageControls _ self makePageControlsFrom: controlSpecs.
	pageControls borderWidth: 0; layoutInset: 0; extent: pageControls width@14.
	pageControls  setProperty: #pageControl toValue: true.
	pageControls setNameTo: 'Page Controls'.
	pageControls eventHandler: (EventHandler new on: #mouseDown send: #move to: self).
	self addMorphBack: pageControls beSticky! !

!MoviePlayerMorph methodsFor: 'access' stamp: 'di 10/13/2000 11:51'!
soundTrackFileName
	^ soundTrackFileName! !

!MoviePlayerMorph methodsFor: 'stepping' stamp: 'dgd 2/22/2003 13:22'!
startRunning
	| ms |
	(frameBufferIfScaled ifNil: [currentPage image]) unhibernate.
	movieFile := AsyncFile new 
				open: (FileDirectory default fullNameFor: movieFileName)
				forWrite: false.
	movieFile 
		primReadStart: movieFile fileHandle
		fPosition: (self filePosForFrameNo: frameNumber)
		count: self fileByteCountPerFrame.
	scorePlayer isNil 
		ifTrue: 
			[ms := Time millisecondClockValue.
			msAtStart := ms - ((frameNumber - 1) * msPerFrame).
			msAtLastSync := ms - msAtStart]
		ifFalse: 
			[(playDirection > 0 and: [scorePlayer isKindOf: SampledSound]) 
				ifTrue: 
					[scorePlayer
						reset;
						playSilentlyUntil: (frameNumber - 1) * msPerFrame / 1000.0;
						initialVolume: 1.0.
					
					[scorePlayer resumePlaying.
					msAtLastSync := scorePlayer millisecondsSinceStart] 
							forkAt: Processor userInterruptPriority].
			msAtLastSync := scorePlayer millisecondsSinceStart].
	frameAtLastSync := frameNumber! !

!MoviePlayerMorph methodsFor: 'stepping' stamp: 'dgd 2/22/2003 13:22'!
step
	"NOTE:  The movie player has two modes of play, depending on whether scorePlayer is nil or not.  If scorePlayer is nil, then play runs according to the millisecond clock.  If scorePlayer is not nil, then the scorePlayer is consulted for synchronization.  If the movie is running ahead, then some calls on step will skip their  action until the right time.  If the movie is running behind, then the frame may advance by more than one to maintain synchronization."

	"ALSO: This player operates with overlapped disk i/o.  This means that while one frame is being displayed, the next frame in sequence is being read into a disk buffer.  The value of frameNumber corresponds to the frame currently visible."

	"This code may not work right for playing backwards right now.
	Single-step and backwards (dir <= 0) should just run open-loop."

	| byteCount simTime ms nextFrameNumber |
	movieFile isNil ifTrue: [^self].
	scorePlayer isNil 
		ifTrue: 
			[(ms := Time millisecondClockValue) < msAtStart 
				ifTrue: 
					["clock rollover"

					msAtStart := ms - (frameNumber * msPerFrame)].
			simTime := ms - msAtStart]
		ifFalse: [simTime := scorePlayer millisecondsSinceStart].
	playDirection > 0 
		ifTrue: 
			[nextFrameNumber := frameAtLastSync 
						+ ((simTime - msAtLastSync) // msPerFrame).
			nextFrameNumber = frameNumber 
				ifTrue: 
					[((scorePlayer isKindOf: AbstractSound) and: [scorePlayer isPlaying not]) 
						ifTrue: [^self stopRunning].
					^self]]
		ifFalse: 
			[nextFrameNumber := playDirection < 0 
						ifTrue: [frameNumber - 1]
						ifFalse: [frameNumber]].
	byteCount := self fileByteCountPerFrame.
	self stepSoundTrack.
	movieFile waitForCompletion.
	movieFile 
		primReadResult: movieFile fileHandle
		intoBuffer: (frameBufferIfScaled ifNil: [currentPage image]) bits
		at: 1
		count: byteCount // 4.
	frameBufferIfScaled ifNotNil: 
			["If this player has been shrunk, then we have to warp to the current page."

			(WarpBlt current toForm: currentPage image)
				sourceForm: frameBufferIfScaled;
				combinationRule: 3;
				cellSize: (playDirection = 0 
							ifTrue: 
								["Use smoothing if just stepping"

								2]
							ifFalse: [1]);
				copyQuad: frameBufferIfScaled boundingBox innerCorners
					toRect: currentPage image boundingBox].
	currentPage changed.
	frameNumber := nextFrameNumber.
	(playDirection = 0 or: 
			[(playDirection > 0 and: [frameNumber >= frameCount]) 
				or: [playDirection < 0 and: [frameNumber <= 1]]]) 
		ifTrue: [^self stopRunning].

	"Start the read operation for the next frame..."
	movieFile 
		primReadStart: movieFile fileHandle
		fPosition: (self filePosForFrameNo: frameNumber)
		count: byteCount! !

!MoviePlayerMorph methodsFor: 'stepping' stamp: 'aoy 2/15/2003 21:45'!
stepSoundTrack
	| x image timeInMillisecs |
	scorePlayer ifNil: [^self].
	soundTrackForm ifNil: [^self].
	timeInMillisecs := playDirection = 0 
		ifTrue: 
			["Stepping forward or back"

			 (frameNumber - 1) * msPerFrame]
		ifFalse: 
			["Driven by sound track"

			 scorePlayer millisecondsSinceStart].
	x := timeInMillisecs / 1000.0 * scorePlayer originalSamplingRate // 250.
	image := soundTrackMorph image.
	image 
		copy: (image boundingBox translateBy: (x - (image width // 2)) @ 0)
		from: soundTrackForm
		to: 0 @ 0
		rule: Form over.
	soundTrackMorph changed! !

!MoviePlayerMorph methodsFor: 'testing' stamp: 'di 8/8/1998 11:57'!
stepTime
	^ 0  "step as fast as possible"! !

!MoviePlayerMorph methodsFor: 'player control' stamp: 'di 10/13/2000 20:06'!
stopPlay

	playDirection _ 0.
	self step! !

!MoviePlayerMorph methodsFor: 'stepping' stamp: 'di 10/8/2000 22:27'!
stopRunning
	"Must only be called with no outstanding file read requests..."
	movieFile ifNotNil: [movieFile close.  movieFile _ nil].
	playDirection _ 0.
	self stopSoundTrackIfAny
! !

!MoviePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/22/2003 13:22'!
stopSoundTrackIfAny
	scorePlayer isNil ifTrue: [^self].
	(scorePlayer isKindOf: SampledSound) 
		ifTrue: [scorePlayer endGracefully]
		ifFalse: [scorePlayer := nil]! !

!MoviePlayerMorph methodsFor: 'menu' stamp: 'di 10/16/2000 13:31'!
thumbnailForThisPage
	"Overridden to make a MovieFrameSyncMorph"

	^ self makeThumbnailInHand: self activeHand
! !

!MoviePlayerMorph methodsFor: 'private' stamp: 'di 10/22/2000 10:39'!
tryToShareScoreFor: fileName

	scorePlayer _ nil.
	self class allInstancesDo:
		[:mp | mp == self ifFalse:
			[mp soundTrackFileName = fileName ifTrue:
				["Found this score already open in another player
					-- return a copy that shares the same sound buffer."
				mp scorePlayer ifNotNil: [^ scorePlayer _ mp scorePlayer copy reset]]]]! !

!MoviePlayerMorph methodsFor: 'rounding' stamp: 'di 10/22/2000 23:41'!
wantsRoundedCorners
	^ Preferences roundedWindowCorners or: [super wantsRoundedCorners]! !

Morph subclass: #MPEGDisplayMorph
	instanceVariableNames: 'frameBuffer mpegFile running desiredFrameRate allowFrameDropping repeat soundTrack volume startMSecs startFrame stopFrame subtitles fullScreen subtitlesDisplayer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Player'!

!MPEGDisplayMorph commentStamp: '<historical>' prior: 0!
I am a simple display screen for an MPEG movie player. My step method advances the movie according to the current frame rate. If necessary, frames as skipped to maintain the desired frame rate. However, since even skipping frames takes time, it may not be possible to achieve fast frame rates with large frame sizes on slow machines.
!

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'jm 12/14/2001 15:34'!
addSoundtrack
	"Add a soundtrack to this JPEG movie."

	| result soundFileName menu compression |
	(mpegFile isKindOf: JPEGMovieFile) ifFalse: [^ self].  "do nothing if not a JPEG movie"

	result _ StandardFileMenu oldFile.
	result ifNil: [^ self].
	soundFileName _ result directory pathName, FileDirectory slash, result name.

	menu _ CustomMenu new title: 'Compression type:'.
	menu addList: #(
		('none (353 kbps)' none)
		('mulaw (176 kbps)' mulaw)
		('adpcm5 (110 kbps)' adpcm5)
		('adpcm4 (88 kbps)' adpcm4)
		('adpcm3 (66 kbps)' adpcm3)
		('gsm (36 kbps)' gsm)).
	compression _ menu startUp.
	compression ifNil: [^ self].

	mpegFile closeFile.
	JPEGMovieFile
		addSoundtrack: soundFileName
		toJPEGMovieNamed: mpegFile fileName
		compressionType: compression.
	self openFileNamed: mpegFile fileName.
! !

!MPEGDisplayMorph methodsFor: 'other' stamp: 'jm 10/11/2004 00:20'!
advanceFrame
	"Advance to the next frame if it is time to do so, skipping frames if necessary."

	| msecs currentFrame desiredFrame framesToAdvance |
	mpegFile hasVideo ifFalse: [^ self].
	soundTrack
		ifNil: [msecs _ Time millisecondClockValue - startMSecs]
		ifNotNil: [msecs _ soundTrack millisecondsSinceStart - SoundPlayer bufferMSecs].
	desiredFrame _ startFrame + ((msecs * desiredFrameRate) // 1000) + 1.
	desiredFrame _ desiredFrame min: (mpegFile videoFrames: 0).
	currentFrame _ mpegFile videoGetFrame: 0.
	stopFrame ifNotNil:
		[desiredFrame _ desiredFrame min: stopFrame.
		currentFrame >= stopFrame ifTrue: [^ self stopPlaying]].
	framesToAdvance _ desiredFrame - currentFrame.
	framesToAdvance <= 0 ifTrue: [^ self].
	(allowFrameDropping and: [framesToAdvance > 1]) ifTrue:
		[mpegFile videoDropFrames: framesToAdvance - 1 stream: 0].
	self nextFrame! !

!MPEGDisplayMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 12:46'!
advancedSubMenu
	"private - create the advanced submenu"

	| subMenu |

	subMenu := MenuMorph new.
	subMenu defaultTarget: self.

	repeat
		ifTrue: [subMenu add: 'turn off repeat (now on)' translated action: #toggleRepeat]
		ifFalse: [subMenu add: 'turn on repeat (now off)' translated action: #toggleRepeat].

	subMenu addLine.

	subMenu addList:	{
		{'set frame rate' translated.									#setFrameRate}.
		#-.
		{'create JPEG movie from MPEG' translated.				#createJPEGfromMPEG}.
		{'create JPEG movie from SqueakMovie' translated.		#createJPEGfromSqueakMovie}.
		{'create JPEG movie from folder of frames' translated.	#createJPEGfromFolderOfFrames}
	}.

	(mpegFile isKindOf: JPEGMovieFile) ifTrue: [
		subMenu addLine.
		mpegFile hasAudio
			ifTrue: [subMenu add: 'remove all soundtracks' translated action: #removeAllSoundtracks]
			ifFalse: [subMenu add: 'add soundtrack' translated action: #addSoundtrack]].

	^ subMenu
! !

!MPEGDisplayMorph methodsFor: 'drawing' stamp: 'jm 3/20/2001 15:57'!
areasRemainingToFill: aRectangle
	"Drawing optimization. Since I completely fill my bounds with opaque pixels, this method tells Morphic that it isn't necessary to draw any morphs covered by me."
	
	^ aRectangle areasOutside: self bounds
! !

!MPEGDisplayMorph methodsFor: 'file open/close' stamp: 'dgd 3/8/2004 22:09'!
closeFile
	"Close my MPEG file, if any."

	mpegFile isNil
		ifFalse: [
			mpegFile closeFile.
			mpegFile := nil.
			frameBuffer := nil].

	subtitles := nil.
	self changed.
! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'yo 2/20/2005 20:10'!
createJPEGfromFolderOfFrames
	"Create a new JPEG movie file from an folder of individual frames. Prompt the user for the folder and file names and the quality setting, then do the conversion."

	| result folderName jpegFileName q frameRate |
	result := StandardFileMenu oldFile.
	result ifNil: [^self].
	folderName := result directory pathName.
	jpegFileName := FillInTheBlank request: 'New movie name?' translated.
	jpegFileName isEmpty ifTrue: [^self].
	(jpegFileName asLowercase endsWith: '.jmv') 
		ifFalse: [jpegFileName := jpegFileName , '.jmv'].
	result := FillInTheBlank request: 'Quality level (1 to 100)?' translated.
	q := result ifNil: [50]
				ifNotNil: [(result asNumber rounded max: 1) min: 100].
	result := FillInTheBlank request: 'Frame rate?' translated.
	frameRate := result ifNil: [10]
				ifNotNil: [(result asNumber rounded max: 1) min: 100].
	JPEGMovieFile 
		convertFromFolderOfFramesNamed: folderName
		toJPEGMovieNamed: jpegFileName
		frameRate: frameRate
		quality: q! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'yo 2/20/2005 20:04'!
createJPEGfromMPEG
	"Create a new JPEG movie file from an MPEG movie. Prompt the user for the file names and the quality setting, then do the conversion."

	| result mpegFileName jpegFileName q |
	result := StandardFileMenu oldFile.
	result ifNil: [^self].
	mpegFileName := result directory pathName , FileDirectory slash 
				, result name.
	jpegFileName := FillInTheBlank request: 'New movie name?' translated.
	jpegFileName isEmpty ifTrue: [^self].
	(jpegFileName asLowercase endsWith: '.jmv') 
		ifFalse: [jpegFileName := jpegFileName , '.jmv'].
	result := FillInTheBlank request: 'Quality level (1 to 100)?' translated.
	q := result ifNil: [50]
				ifNotNil: [(result asNumber rounded max: 1) min: 100].
	JPEGMovieFile 
		convertMPEGFileNamed: mpegFileName
		toJPEGMovieNamed: jpegFileName
		quality: q! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'yo 2/20/2005 20:05'!
createJPEGfromSqueakMovie
	"Create a new JPEG movie file from an SqueakTime movie. Prompt the user for the file names and the quality setting, then do the conversion."

	| result squeakMovieFileName jpegFileName q |
	result := StandardFileMenu oldFile.
	result ifNil: [^self].
	squeakMovieFileName := result directory pathName , FileDirectory slash 
				, result name.
	jpegFileName := FillInTheBlank request: 'New movie name?' translated.
	jpegFileName isEmpty ifTrue: [^self].
	(jpegFileName asLowercase endsWith: '.jmv') 
		ifFalse: [jpegFileName := jpegFileName , '.jmv'].
	result := FillInTheBlank request: 'Quality level (1 to 100)?' translated.
	q := result ifNil: [50]
				ifNotNil: [(result asNumber rounded max: 1) min: 100].
	JPEGMovieFile 
		convertSqueakMovieNamed: squeakMovieFileName
		toJPEGMovieNamed: jpegFileName
		quality: q! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jm 10/12/2004 12:22'!
currentFrameScaled
	"Answer a Form containing the current frame scaled to my current size."

	| f |
	f _ Form extent: self extent depth: 32.
	frameBuffer ifNil: [^ f fillColor: (Color gray: 0.75)].
	self drawScaledOn: ((FormCanvas on: f) copyOffset: self topLeft negated).
	^ f! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'dgd 3/8/2004 13:34'!
doubleSize
	"change the receiver's extent to double of the normal size"
	self magnifyBy: 2! !

!MPEGDisplayMorph methodsFor: 'drawing' stamp: 'jm 11/11/2001 15:49'!
drawOn: aCanvas
	"Draw the current frame image, if there is one. Otherwise, fill screen with gray."

	frameBuffer
		ifNil: [aCanvas fillRectangle: self bounds color: (Color gray: 0.75)]
		ifNotNil: [
			self extent = frameBuffer extent
				ifTrue: [aCanvas drawImage: frameBuffer at: bounds origin]
				ifFalse: [self drawScaledOn: aCanvas]].
! !

!MPEGDisplayMorph methodsFor: 'drawing' stamp: 'jm 11/13/2001 08:45'!
drawScaledOn: aCanvas
	"Draw the current frame image scaled to my bounds."

	| outForm destPoint warpBlt |
	((aCanvas isKindOf: FormCanvas) and: [aCanvas form = Display])
		ifTrue: [  "optimization: when canvas is the Display, Warpblt directly to it"
			outForm _ Display.
			destPoint _ bounds origin + aCanvas origin]
		ifFalse: [
			outForm _ Form extent: self extent depth: aCanvas form depth.
			destPoint _ 0@0].
	warpBlt _ (WarpBlt current toForm: outForm)
		sourceForm: frameBuffer;
		colorMap: (frameBuffer colormapIfNeededForDepth: outForm depth);
		cellSize: 1;  "installs a new colormap if cellSize > 1"
		combinationRule: Form over.
	outForm == Display ifTrue: [warpBlt clipRect: aCanvas clipRect].
	warpBlt
		copyQuad: frameBuffer boundingBox innerCorners
		toRect: (destPoint extent: self extent).
	outForm == Display ifFalse: [
		aCanvas drawImage: outForm at: bounds origin].

! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jm 11/16/2001 16:40'!
extent: aPoint
	"Overridden to maintain movie aspect ratio."

	| scale |
	frameBuffer ifNil: [^ super extent: aPoint].
	scale _ (aPoint x / frameBuffer width) max: (aPoint y / frameBuffer height).
	scale _ scale max: (16 / frameBuffer width).
	super extent: (frameBuffer extent * scale) rounded.
! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 3/8/2004 18:57'!
fullFileName
	"answer the receiver's fullFileName"
	^ mpegFile isNil
		ifTrue: ['']
		ifFalse: [mpegFile fileName]! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 3/7/2004 18:52'!
fullScreen
	"answer whatever the receiver is fullScreen

	Note: comparation with true to make it work with instances created before the introduccion of the variable"
	^ fullScreen == true! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 3/7/2004 18:56'!
fullScreen: aBoolean 
	"change the receiver's fullScreen"
	fullScreen := aBoolean! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'dgd 3/8/2004 13:34'!
halfSize
	"change the receiver's extent to a half of the normal size"
	self magnifyBy: 1 / 2! !

!MPEGDisplayMorph methodsFor: 'event handling' stamp: 'dgd 3/7/2004 20:06'!
handlesKeyboard: evt
	^ true! !

!MPEGDisplayMorph methodsFor: 'event handling' stamp: 'dgd 3/7/2004 23:02'!
handlesMouseDown: evt 
	^ evt yellowButtonPressed! !

!MPEGDisplayMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 22:11'!
hasSubtitles
	"answer if the receiver has subtitles or not"
	^ mpegFile isNil not
		and: [subtitles isNil not]! !

!MPEGDisplayMorph methodsFor: 'initialization' stamp: 'dgd 3/8/2004 23:05'!
initialize
	"initialize the state of the receiver"
	super initialize.""
	super extent: 250 @ 0.
	frameBuffer := nil.
	mpegFile := nil.
	running := false.
	desiredFrameRate := 10.0.
	allowFrameDropping := true.
	repeat := false.
	soundTrack := nil.
	volume := 0.5.
	fullScreen := false.
""
self initializeSubtitlesDisplayer! !

!MPEGDisplayMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 23:03'!
initializeSubtitlesDisplayer
	"private - builds the subtitle displayer"
	subtitlesDisplayer := MPEGSubtitlesDisplayer on: self selector: #subtitle.
subtitlesDisplayer contents:''.
	self addMorphFront: subtitlesDisplayer.
	^ subtitlesDisplayer! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'dgd 3/8/2004 21:05'!
invokeMenu
	"Invoke a menu of additonal functions."

	| aMenu |
	aMenu := MenuMorph new.
	aMenu defaultTarget: self.

	aMenu addList:	{
		{'open file (o)' translated.			#openMPEGFile}.
		#-.
		{'rewind (r)' translated.				#rewindMovie}.
		{'play (p)' translated.				#startPlaying}.
		{'stop (s)' translated.				#stopPlaying}.
		{'previous frame (<-)' translated.	#previousFrame}.
		{'next frame (->)' translated.		#nextFrame}.
		#-.
	}.

	aMenu addLine.
	aMenu add: 'zoom' translated subMenu: self zoomSubMenu.
	aMenu add: 'subtitles' translated subMenu: self subtitlesSubMenu.
	aMenu add: 'advanced' translated subMenu: self advancedSubMenu.

	aMenu popUpEvent: self world activeHand lastEvent in: self world
! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 2/15/2004 21:41'!
isRunning
	"answer whatever the receiver is running"
	^ running! !

!MPEGDisplayMorph methodsFor: 'other' stamp: 'jm 11/14/2001 11:58'!
jpegMovieSize: quality
	"Convert all my frames to a JPEG and measure the total size."

	| jpegSize jpegDecodeTime jpegStream t outForm |
	mpegFile hasVideo ifFalse: [^ self error: 'movie has no video'].
	jpegSize _ 0.
	jpegDecodeTime _ 0.
	jpegStream _ WriteStream on: (ByteArray new: 100000).
	self rewindMovie.

	[(mpegFile videoGetFrame: 0) < (mpegFile videoFrames: 0)] whileTrue: [
		jpegStream reset.
		(JPEGReadWriter2 on: jpegStream)
			nextPutImage: frameBuffer
			quality: quality
			progressiveJPEG: false.
		jpegSize _ jpegSize + jpegStream position.
		t _ [
			outForm _ (JPEGReadWriter2 on: (ReadStream on: jpegStream contents)) nextImage
		] timeToRun.
		jpegDecodeTime _ jpegDecodeTime + t.
		outForm display.
		frameBuffer displayAt: (outForm width + 10)@0.
		self nextFrame].

	^ Array with: jpegSize with: jpegDecodeTime with: (mpegFile videoFrames: 0)

! !

!MPEGDisplayMorph methodsFor: 'event handling' stamp: 'dgd 3/7/2004 22:37'!
keyStroke: evt 
	| char asc |

	char := evt keyCharacter.
	asc := char asciiValue.
	(char = $o or:[ char = $O])
		ifTrue: 
			["open o/O" 
			self openMPEGFile. ^self].
	(char = $m or:[ char = $M])
		ifTrue: 
			["menu key m/M" 
			self invokeMenu. ^self].
	(char = $r or:[ char = $R])
		ifTrue: 
			["rewind r/R" 
			self rewindMovie. ^self].
	(char = $p or:[ char = $P])
		ifTrue: 
			["play p/P" 
			self startPlaying. ^self].
	(char = $s or:[ char = $S])
		ifTrue: 
			["stop s/S" 
			self stopPlaying. ^self].
	(asc = 28)
		ifTrue: 
			[ "left arrow key" 
			self previousFrame. ^self].
	(asc = 29)
		ifTrue: 
			[ "right arrow key" 
			self nextFrame. ^self].
	(char = $u or:[ char = $U])
		ifTrue: 
			["subtitles file u/U" 
			self openSubtitlesFile. ^self].! !

!MPEGDisplayMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 14:12'!
magnifyBy: aNumber 
	"private - scale the video (if any) to a scale of the normalExtent"
	| ne |
	fullScreen := false.""
	ne := self normalExtent.
	ne isNil
		ifFalse: [self extent: (ne * aNumber) rounded]! !

!MPEGDisplayMorph methodsFor: 'other' stamp: 'jm 11/21/2001 16:58'!
measureMaxFrameRate
	"For testing. Play through the movie as fast as possible, updating the world each time, and report the frame rate."

	| oldFrameRate oldFrameDropping t |
	self rewindMovie.
	oldFrameRate _ desiredFrameRate.
	oldFrameDropping _ allowFrameDropping.
	desiredFrameRate _ 1000.0.
	allowFrameDropping _ false.

	self startPlaying.
	t _ [[running] whileTrue: [self world doOneCycleNow]] timeToRun.

	desiredFrameRate _ oldFrameRate.
	allowFrameDropping _ oldFrameDropping.

	^ (mpegFile videoFrames: 0) / (t / 1000.0)
! !

!MPEGDisplayMorph methodsFor: 'event handling' stamp: 'dgd 3/7/2004 22:13'!
mouseDown: evt 
	evt yellowButtonPressed
		ifTrue: [^ self invokeMenu].
	super mouseDown: evt! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'mu 6/25/2003 02:41'!
moviePosition
	"Answer a number between 0.0 and 1.0 indicating the current position within the movie."

	mpegFile ifNil: [^ 0.0].
	mpegFile fileHandle ifNil: [^ 0.0].
	(FileStream isAFileNamed: mpegFile fileName) ifFalse: [^0.0].
	mpegFile hasVideo
		ifTrue: [^ ((mpegFile videoGetFrame: 0) asFloat / (mpegFile videoFrames: 0)) min: 1.0].
	soundTrack ifNotNil: [^ soundTrack soundPosition].
	^ 0.0
! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jm 12/16/2001 12:34'!
moviePosition: fraction
	"Jump to the position the given fraction through the movie. The argument is a number between 0.0 and 1.0."

	| frameCount frameIndex |
	self mpegFileIsOpen ifFalse: [^ self].
	self stopPlaying.
	mpegFile hasVideo ifTrue: [
		frameCount _ mpegFile videoFrames: 0.
		frameIndex _ (frameCount * fraction) truncated - 1.
		frameIndex _ (frameIndex max: 0) min: (frameCount - 3).
		mpegFile videoSetFrame: frameIndex stream: 0.
		^ self nextFrame].

	mpegFile hasAudio ifTrue: [
		soundTrack soundPosition: fraction].
! !

!MPEGDisplayMorph methodsFor: 'file open/close' stamp: 'jm 3/22/2001 12:42'!
mpegFileIsOpen
	"Answer true if I have an open, valid MPEG file handle. If the handle is not valid, try to re-open the file."

	mpegFile ifNil: [^ false].
	mpegFile fileHandle ifNil: [
		"try to reopen the file, which may have been saved in a snapshot"
		mpegFile openFile: mpegFile fileName.
		mpegFile fileHandle ifNil: [mpegFile _ nil]].
	^ mpegFile notNil
! !

!MPEGDisplayMorph methodsFor: 'commands' stamp: 'jm 11/16/2001 15:39'!
nextFrame
	"Fetch the next frame into the frame buffer."

	mpegFile ifNil: [^ self].
	mpegFile videoReadFrameInto: frameBuffer stream: 0.
	self changed.
! !

!MPEGDisplayMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 13:30'!
normalExtent
	"private - answer the extent of the video, if any"
	(mpegFile isNil
			or: [mpegFile hasVideo not])
		ifTrue: [^ nil].
	""
	^ (mpegFile videoFrameWidth: 0)
		@ (mpegFile videoFrameHeight: 0)! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'dgd 3/8/2004 13:34'!
normalSize
	"change the receiver's extent to the normal size"
	self magnifyBy: 1! !

!MPEGDisplayMorph methodsFor: 'file open/close' stamp: 'dgd 3/7/2004 19:07'!
openFileNamed: mpegFileName
	"Try to open the MPEG file with the given name. Answer true if successful."

	| e |
	self closeFile.
	(FileDirectory default fileExists: mpegFileName)
		ifFalse: [self inform: ('File not found: {1}' translated format: {mpegFileName}). ^ false].

	(MPEGFile isFileValidMPEG: mpegFileName)
		ifTrue: [mpegFile := MPEGFile openFile: mpegFileName]
		ifFalse: [
			(JPEGMovieFile isJPEGMovieFile: mpegFileName)
				ifTrue: [mpegFile := JPEGMovieFile new openFileNamed: mpegFileName]
				ifFalse: [self inform: ('Not an MPEG or JPEG movie file: {1}' translated format: {mpegFileName}). ^ false]].
	mpegFile fileHandle ifNil: [^ false].

	"initialize soundTrack"
	mpegFile hasAudio
		ifTrue: [soundTrack := mpegFile audioPlayerForChannel: 1]
		ifFalse: [soundTrack := nil].

	mpegFile hasVideo
		ifTrue: [  "set screen size and display first frame"
			desiredFrameRate := mpegFile videoFrameRate: 0.
			soundTrack ifNotNil: [  "compute frame rate from length of audio track"
				desiredFrameRate := (mpegFile videoFrames: 0) / soundTrack duration].
			e := (mpegFile videoFrameWidth: 0)@(mpegFile videoFrameHeight: 0).
			frameBuffer := Form extent: e depth: (Display depth max: 16).
			super extent: e.
			self nextFrame]
		ifFalse: [  "hide screen for audio-only files"
			super extent: 250@0].
! !

!MPEGDisplayMorph methodsFor: 'file open/close' stamp: 'jm 11/26/2001 08:26'!
openMPEGFile
	"Invoked by the 'Open' button. Prompt for a file name and try to open that file as an MPEG file."

	| result |
	result _ (FileList2 modalFileSelectorForSuffixes: #('mp3' 'mpg' 'mpeg' 'jmv')) .
	result ifNil: [^ self].
	self stopPlaying.
	self openFileNamed: (result fullName).
! !

!MPEGDisplayMorph methodsFor: 'file open/close' stamp: 'dgd 3/8/2004 20:16'!
openSubtitlesFile
	"Invoked by the 'Subtitles' button. Prompt for a file name and 
	try to open that file as a subs file."
	| result |
	self mpegFileIsOpen
		ifFalse: [^ self].
	mpegFile hasVideo
		ifFalse: [self inform: 'select a video file' translated.
			^ self].
	result := FileList2 modalFileSelectorForSuffixes: #('sub' ).
	result
		ifNil: [^ self].
	self openSubtitlesFileNamed: result fullName! !

!MPEGDisplayMorph methodsFor: 'file open/close' stamp: 'dgd 3/8/2004 22:58'!
openSubtitlesFileNamed: aString 
	"Try to open the subtitle file with the given name. Answer true  
	if successful."
	subtitles := nil.
	""
	"try to create the displayer.  it's useful for instances of mpegplayer older than the subtitles support"
	self subtitlesDisplayer.
	""
	(FileDirectory default fileExists: aString)
		ifFalse: [self
				inform: ('File not found: {1}' translated format: {aString}).
			^ false].
	Utilities
		informUser: 'opening the file, please wait' translated
		during: [subtitles := MPEGSubtitles fromFileNamed: aString]! !

!MPEGDisplayMorph methodsFor: 'commands' stamp: 'jm 10/11/2004 23:06'!
playUntilPosition: finalPosition
	"Play the movie until the given position, then stop."

	| totalFrames |
	totalFrames _ self totalFrames.
	(totalFrames > 0 and: [finalPosition > 0]) ifFalse: [^ self].  "do nothing"
	self startPlaying.
	stopFrame _ (finalPosition * totalFrames) asInteger min: totalFrames! !

!MPEGDisplayMorph methodsFor: 'commands' stamp: 'jm 4/6/2001 08:31'!
previousFrame
	"Go to the previous frame."

	| n |
	mpegFile ifNil: [^ self].
	running ifTrue: [^ self].
	n _ (mpegFile videoGetFrame: 0) - 2.
	n _ (n min: ((mpegFile videoFrames: 0) - 3)) max: 0.
	mpegFile videoSetFrame: n stream: 0.
	self nextFrame.
! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'jm 12/13/2001 20:48'!
removeAllSoundtracks
	"Remove all soundtracks from this JPEG movie."

	(mpegFile isKindOf: JPEGMovieFile) ifFalse: [^ self].  "do nothing if not a JPEG movie"

	mpegFile closeFile.
	JPEGMovieFile removeSoundtrackFromJPEGMovieNamed: mpegFile fileName.
	self openFileNamed: mpegFile fileName.
! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jm 6/3/2001 14:34'!
repeat
	"Answer the repeat flag."

	repeat ifNil: [repeat _ false].
	^ repeat
! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jm 6/3/2001 14:33'!
repeat: aBoolean
	"Set the repeat flag. If true, the movie will loop back to the beginning when it gets to the end."

	repeat _ aBoolean.
! !

!MPEGDisplayMorph methodsFor: 'commands' stamp: 'jm 11/13/2001 07:36'!
rewindMovie
	"Rewind to the beginning of the movie."
	"Details: Seeking by percent or frame number both seem to have problems, so just re-open the file."

	| savedExtent savedRate |
	self mpegFileIsOpen ifFalse: [^ self].
	self stopPlaying.

	"re-open the movie, retaining current extent and frame rate"
	savedExtent _ self extent.
	savedRate _ desiredFrameRate.
	self openFileNamed: mpegFile fileName.  "recomputes rate and extent"
	self extent: savedExtent.
	desiredFrameRate _ savedRate.
! !

!MPEGDisplayMorph methodsFor: 'commands' stamp: 'dgd 10/8/2003 19:10'!
setFrameRate
	"Ask the user to specify the desired frame rate."

	| rateString |
	rateString := FillInTheBlank request: 'Desired frames per second?' translated
				initialAnswer: desiredFrameRate printString.
	rateString isEmpty ifTrue: [^self].
	desiredFrameRate := rateString asNumber asFloat.
	desiredFrameRate := desiredFrameRate max: 0.1! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'dgd 3/8/2004 22:57'!
setSubtitlesBackgroundColor
	"open a dialog to change the background color of the subtitles"
	self subtitlesDisplayer openAPropertySheet! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'dgd 3/8/2004 22:57'!
setSubtitlesColor
	"open a dialog to change the color of the subtitles"
	self subtitlesDisplayer changeSubtitlesColor! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'dgd 3/8/2004 22:57'!
setSubtitlesFont
	"change the subtitles font"
	self subtitlesDisplayer changeFont! !

!MPEGDisplayMorph methodsFor: 'commands' stamp: 'jm 10/10/2004 23:58'!
startPlaying
	"Start playing the movie at the current position."

	| frameIndex |
	self stopPlaying.
	stopFrame _ nil.
	self mpegFileIsOpen ifFalse: [^ self].

	(FileStream isAFileNamed: mpegFile fileName) ifFalse: [ | newFileResult newFileName |
		self inform: 'Path changed. Enter new one for: ', (FileDirectory localNameFor: mpegFile fileName).
		newFileResult _ StandardFileMenu oldFile.
		newFileName _ newFileResult directory fullNameFor: newFileResult name.	
		mpegFile openFile: newFileName].
	
	mpegFile hasAudio
		ifTrue:
			[mpegFile hasVideo ifTrue:
				["set movie frame position from soundTrack position"
				soundTrack reset.  "ensure file is open before positioning"
				soundTrack soundPosition: (mpegFile videoGetFrame: 0) asFloat / (mpegFile videoFrames: 0).
				"now set frame index from the soundtrack position for best sync"
				frameIndex _ ((soundTrack millisecondsSinceStart * desiredFrameRate) // 1000).
				frameIndex _ (frameIndex max: 0) min: ((mpegFile videoFrames: 0) - 3).
				mpegFile videoSetFrame: frameIndex stream: 0].

			SoundPlayer stopReverb.
			soundTrack volume: volume.
			soundTrack repeat: repeat.
			soundTrack resumePlaying.
			startFrame _ startMSecs _ 0]
		ifFalse:
			[soundTrack _ nil.
			startFrame _ mpegFile videoGetFrame: 0.
			startMSecs _ Time millisecondClockValue].
	running _ true! !

!MPEGDisplayMorph methodsFor: 'stepping' stamp: 'jm 6/3/2001 18:38'!
step
	"If I'm running and the mpegFile is open and has video, advance to the next frame. Stop if we we hit the end of the video."

	running ifFalse: [^ self].
	mpegFile ifNil: [^ self].

	(mpegFile hasVideo and:
	 [(mpegFile videoGetFrame: 0) >= (mpegFile videoFrames: 0)])
		ifTrue: [  "end of video"
			self stopPlaying.
			repeat ifTrue: [
				self rewindMovie.
				self startPlaying]]
		ifFalse: [self advanceFrame].
! !

!MPEGDisplayMorph methodsFor: 'stepping' stamp: 'jm 4/6/2001 08:47'!
stepTime
	"Run my step method as often as possible. Step does very little work if it is not time to advance to the next frame."

	^ 0
! !

!MPEGDisplayMorph methodsFor: 'commands' stamp: 'jm 6/3/2001 14:30'!
stopPlaying
	"Stop playing the movie."

	running _ false.
	soundTrack ifNotNil: [soundTrack pause].
! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 3/8/2004 22:13'!
subtitle
	"answer the subtitle for the current frame"
	self hasSubtitles
		ifFalse: [^ ''].
	self mpegFileIsOpen
		ifFalse: [^ ''].
	mpegFile hasVideo
		ifFalse:[^''].
""
^ subtitles
				subtitleForFrame: (mpegFile videoGetFrame: 0)! !

!MPEGDisplayMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 22:59'!
subtitlesDisplayer
	"private - answer the receiver's subtitlesDisplayer. create one  
	if needed"
	^ subtitlesDisplayer
		ifNil: [self initializeSubtitlesDisplayer]! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 3/8/2004 22:12'!
subtitlesFileShortName
	"answer the receiver's subtitlesFileShortName"
	| fileFull defaultDirFull fileShort |
	self hasSubtitles ifFalse:[^ ''].
	" 
	answer the shortest path to the file to make easier to move  
	morphs with references to files between different platforms"
	fileFull := subtitles fileName.
	""
	defaultDirFull := FileDirectory default fullName.
	fileShort := (fileFull beginsWith: defaultDirFull)
				ifTrue: [fileFull allButFirst: defaultDirFull size + 1]
				ifFalse: [fileFull].
	""
	^ fileShort! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 3/8/2004 22:44'!
subtitlesFileShortName: aString 
	"change the receiver's subtitlesFileShortName, that means 
	open the subtitles file named aString"
	| fullName |
	self mpegFileIsOpen
		ifFalse: [^ self].
	mpegFile hasVideo
		ifFalse: [^ self].
	""
	fullName := FileDirectory default fullNameFor: aString.
	self openSubtitlesFileNamed: fullName! !

!MPEGDisplayMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 22:22'!
subtitlesSubMenu
	"private - create the subtitles submenu"

	| subMenu |

	subMenu := MenuMorph new.
	subMenu defaultTarget: self.
	subMenu add: 'open subtitles file (u)' translated action: #openSubtitlesFile.

	self hasSubtitles
		ifTrue: [
			subMenu addLine.
			subMenu add: 'set subtitles font' translated action: #setSubtitlesFont.
			subMenu add: 'set subtitles color' translated action: #setSubtitlesColor.
			subMenu add: 'set subtitles background color' translated action: #setSubtitlesBackgroundColor].

	^ subMenu
! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'dgd 3/8/2004 14:13'!
toggleFullScreen
	"Toggle the fullScreen flag."
	mpegFile isNil
		ifTrue: [^ self].
	mpegFile hasVideo
		ifFalse: [^ self].
	""
	self fullScreen: self fullScreen not.
	""
	"set screen size"
	self fullScreen
		ifTrue: [""
			self extent: Display extent.
			World activeHand newMouseFocus: self.
			self comeToFront]
		ifFalse: [self extent: self normalExtent].
	""
	(self fullScreen
			and: [self owner isKindOf: MPEGMoviePlayerMorph])
		ifTrue: [self owner position: -6 @ -6]
		ifFalse: [self owner == self world
				ifFalse: [self owner position: 0 @ 0] ifTrue:[self position:0@0]].
	""
	self nextFrame! !

!MPEGDisplayMorph methodsFor: 'menu' stamp: 'jm 12/13/2001 08:55'!
toggleRepeat
	"Toggle the repeat flag."

	repeat _ repeat not.
! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jm 10/12/2004 11:40'!
totalFrames
	"Answer the total number of frames in this movie."

	mpegFile ifNil: [^ 0].
	mpegFile fileHandle ifNil: [^ 0].
	(FileStream isAFileNamed: mpegFile fileName) ifFalse: [^ 0].
	mpegFile hasVideo ifFalse: [^ 0].
	^ mpegFile videoFrames: 0! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jm 10/12/2004 11:40'!
totalSeconds
	"Answer the total number of seconds in this movie."

	mpegFile ifNil: [^ 0].
	mpegFile fileHandle ifNil: [^ 0].
	(FileStream isAFileNamed: mpegFile fileName) ifFalse: [^ 0].
	mpegFile hasVideo ifFalse: [^ 0].
	^ self totalFrames asFloat / (mpegFile videoFrameRate: 0)! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 3/8/2004 21:47'!
videoFileShortName
	"answer the receiver's videoFileShortName"
	| fileFull defaultDirFull fileShort |
	mpegFile isNil
		ifTrue: [^ ''].
	" 
	answer the shortest path to the file to make easier to move  
	morphs with references to files between different platforms"
	fileFull := mpegFile fileName.
	""
	defaultDirFull := FileDirectory default fullName.
	fileShort := (fileFull beginsWith: defaultDirFull)
				ifTrue: [fileFull allButFirst: defaultDirFull size + 1]
				ifFalse: [fileFull].
	""
	^ fileShort! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'dgd 3/8/2004 22:45'!
videoFileShortName: aString 
	"change the receiver's videoFileShortName, that means open 
	the video file named aString"
	| fullName |
	self stopPlaying.
	fullName := FileDirectory default fullNameFor: aString.
	self openFileNamed: fullName! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jm 6/3/2001 14:35'!
volume
	"Answer the sound playback volume."

	^ volume
! !

!MPEGDisplayMorph methodsFor: 'accessing' stamp: 'jdl 3/28/2003 09:41'!
volume: aNumber 
	"Set the sound playback volume to the given level, between 0.0 and 1.0."

	volume := aNumber asFloat.
	volume := volume max: 0.0.
	volume := volume min: 1.0.
	soundTrack ifNotNil: [soundTrack volume: volume]! !

!MPEGDisplayMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 21:08'!
zoomSubMenu
	"private - create the zoom submenu"

	| subMenu |

	subMenu := MenuMorph new.
	subMenu defaultTarget: self.

	self fullScreen
		ifTrue: [subMenu add: 'turn off full screen' translated action: #toggleFullScreen]
		ifFalse: [subMenu add: 'turn on full screen' translated action: #toggleFullScreen].

	subMenu addLine.
	subMenu add: '50%' action: #halfSize.
	subMenu add: '100%' action: #normalSize.
	subMenu add: '200%' action: #doubleSize.

	^ subMenu
! !

AbstractSound subclass: #StreamingMP3Sound
	instanceVariableNames: 'volume repeat mpegFile mpegStreamIndex totalSamples streamSamplingRate mixer lastBufferMSecs mutex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Kernel'!

!StreamingMP3Sound commentStamp: '<historical>' prior: 0!
I implement a streaming player for MPEG or MP3 files.
Example of use:
	(StreamingMP3Sound onFileNamed: 'song.mp3') play.
!

!StreamingMP3Sound class methodsFor: 'instance creation' stamp: 'jm 11/20/2001 16:35'!
onFileNamed: fileName
	"Answer an instance of me for playing the sound track of the MPEG or MP3 file with the given name. Answer nil the file is not a valid MPEG or MP3 file."

	| mpegFile |
	(MPEGFile isFileValidMPEG: fileName) ifFalse: [^ nil].
	mpegFile _ MPEGFile openFile: fileName.
	^ self new initMPEGFile: mpegFile streamIndex: 0  "assume sound track is in stream 0"
! !

!StreamingMP3Sound methodsFor: 'file ops' stamp: 'jm 11/25/2001 14:11'!
closeFile
	"Close the MP3 or MPEG file."

	self pause.
	mpegFile ifNil: [^ self].
	mpegFile closeFile.
	mpegFile _ nil.
	mixer _ nil.
! !

!StreamingMP3Sound methodsFor: 'private' stamp: 'jm 11/16/2001 15:59'!
createMixer
	"Create a mixed sound consisting of sampled sounds with one sound buffer's worth of samples. The sound has the same sampling rate and number of channels as the MPEG or MP3 file."

	| channels pan snd |
	mpegFile ifNil: [^ self error: 'No MPEG or MP3 file'].
	channels _ mpegFile audioChannels: mpegStreamIndex.
	streamSamplingRate _ mpegFile audioSampleRate: mpegStreamIndex.
	mixer _ MixedSound new.
	1 to: channels do: [:c |
		channels = 1
			ifTrue: [pan _ 0.5]
			ifFalse: [pan _ (c - 1) asFloat / (channels - 1)].
		snd _ SampledSound
			samples: (SoundBuffer newMonoSampleCount: 2)  "buffer size will be adjusted dynamically"
			samplingRate: streamSamplingRate.
		mixer add: snd pan: pan volume: volume].
! !

!StreamingMP3Sound methodsFor: 'accessing' stamp: 'jm 11/16/2001 17:16'!
duration
	"Answer the duration of this sound in seconds."

	^ totalSamples asFloat / streamSamplingRate
! !

!StreamingMP3Sound methodsFor: 'initialization' stamp: 'jm 11/27/2001 10:06'!
initMPEGFile: anMPEGFile streamIndex: anInteger
	"Initialize for playing the given stream of the given MPEG or MP3 file."

	volume _ 0.3.
	repeat _ false.
	mpegFile _ anMPEGFile.
	mpegStreamIndex _ anInteger.
	totalSamples _ mpegFile audioSamples: mpegStreamIndex.
	self reset.
! !

!StreamingMP3Sound methodsFor: 'private' stamp: 'crl 1/23/2003 14:26'!
loadBuffersForSampleCount: count
        "Load the sound buffers for all tracks with the next count
samples from the MPEG
file sound track."

        | snd buf |
        1 to: mixer sounds size do: [:i |
                snd _ mixer sounds at: i.
                buf _ snd samples.
                buf monoSampleCount = count ifFalse: [
                        buf _ SoundBuffer newMonoSampleCount: count.
                        snd setSamples: buf samplingRate:
streamSamplingRate].
                i = 1 ifTrue: [  "first channel"
                                mpegFile
                                        audioReadBuffer: buf
                                        stream: mpegStreamIndex
                                        channel: 0]
                        ifFalse: [  "all other channels"
                                mpegFile
                                        audioReReadBuffer: buf
                                        stream: mpegStreamIndex
                                        channel: 1]].
        mixer reset.
! !

!StreamingMP3Sound methodsFor: 'playing' stamp: 'jm 11/27/2001 10:16'!
millisecondsSinceStart
	"Answer the number of milliseconds since this sound started playing."

	| i mSecs |
	mpegFile ifNil: [^ 0].
	mpegFile fileHandle ifNil: [^ 0].  "mpeg file not open"
	i _ mpegFile audioGetSample: mpegStreamIndex.
	i < 0 ifTrue: [^ 0].  "movie file has no audio"
	mSecs _ i * 1000 // streamSamplingRate.
	(self isPlaying and: [lastBufferMSecs > 0]) ifTrue: [
		"adjust mSecs by the milliseconds since the last buffer"
		mutex critical: [
			mSecs _ i * 1000 // streamSamplingRate.
			mSecs _ mSecs + ((Time millisecondClockValue - lastBufferMSecs) max: 0)]].
	^ mSecs + 350 - (2 * SoundPlayer bufferMSecs)
! !

!StreamingMP3Sound methodsFor: 'file ops' stamp: 'jm 11/16/2001 15:58'!
mpegFileIsOpen
	"Answer true if I have an open, valid MPEG file handle. If the handle is not valid, try to re-open the file."

	mpegFile ifNil: [^ false].
	mpegFile fileHandle ifNil: [
		"try to reopen the file, which may have been saved in a snapshot"
		mpegFile openFile: mpegFile fileName.
		mpegFile fileHandle ifNil: [mpegFile _ nil]].
	^ mpegFile notNil
! !

!StreamingMP3Sound methodsFor: 'playing' stamp: 'jm 11/27/2001 10:09'!
playSampleCount: n into: aSoundBuffer startingAt: startIndex
	"Mix the next n samples of this sound into the given buffer starting at the given index"

	| current |
	self repeat ifTrue: [  "loop if necessary"
		current _ mpegFile audioGetSample: mpegStreamIndex.
		(totalSamples - current) < n ifTrue: [
			mpegFile audioSetSample: 0 stream: mpegStreamIndex]].

	mutex critical: [
		lastBufferMSecs _ Time millisecondClockValue.
		self loadBuffersForSampleCount: (n * streamSamplingRate) // SoundPlayer samplingRate.
		mixer playSampleCount: n into: aSoundBuffer startingAt: startIndex].
! !

!StreamingMP3Sound methodsFor: 'accessing' stamp: 'jm 6/3/2001 15:43'!
repeat
	"Answer the repeat flag."

	repeat ifNil: [repeat _ false].
	^ repeat
! !

!StreamingMP3Sound methodsFor: 'accessing' stamp: 'jm 6/3/2001 18:39'!
repeat: aBoolean
	"Set the repeat flag. If true, this sound will loop back to the beginning when it gets to the end."

	repeat _ aBoolean.
! !

!StreamingMP3Sound methodsFor: 'playing' stamp: 'jm 11/27/2001 10:06'!
reset

	super reset.
	self createMixer.
	mpegFile audioSetSample: 0 stream: mpegStreamIndex.
	lastBufferMSecs _ 0.
	mutex _ Semaphore forMutualExclusion.
! !

!StreamingMP3Sound methodsFor: 'playing' stamp: 'jm 11/16/2001 15:59'!
samplesRemaining

	| samplesPlayed |
	mpegFile ifNil: [^ 0].
	self repeat ifTrue: [^ 1000000].
	samplesPlayed _ mpegFile audioGetSample: mpegStreamIndex.
	samplesPlayed > totalSamples ifTrue: [^ 0].
	^ totalSamples - samplesPlayed
! !

!StreamingMP3Sound methodsFor: 'converting' stamp: 'jm 12/13/2001 20:32'!
saveAsFileNamed: newFileName compressionType: compressionTypeString
	"Store this MP3 sound in a SunAudio file with the given name using the given compression type."

	| outFile |
	outFile _ (FileStream newFileNamed: newFileName) binary.
	self storeSunAudioOn: outFile compressionType: compressionTypeString.
	outFile close.
! !

!StreamingMP3Sound methodsFor: 'accessing' stamp: 'jm 11/16/2001 16:18'!
soundPosition
	"Answer the relative position of sound playback as a number between 0.0 and 1.0."

	self mpegFileIsOpen ifFalse: [^ 0.0].
	mpegFile hasAudio ifFalse: [^ 0.0].
	^ (mpegFile audioGetSample: 0) asFloat / totalSamples
! !

!StreamingMP3Sound methodsFor: 'accessing' stamp: 'jm 11/16/2001 16:19'!
soundPosition: fraction
	"Jump to the position the given fraction through the sound file. The argument is a number between 0.0 and 1.0."

	| sampleIndex |
	self mpegFileIsOpen ifFalse: [^ self].
	mpegFile hasAudio ifTrue: [
		sampleIndex _ ((totalSamples * fraction) truncated max: 0) min: totalSamples.
		mpegFile audioSetSample: 0 stream: 0.  "work around for library bug: first seek to zero"
		mpegFile audioSetSample: sampleIndex stream: 0].
! !

!StreamingMP3Sound methodsFor: 'converting' stamp: 'jm 12/14/2001 12:39'!
storeSunAudioOn: aBinaryStream compressionType: compressionName
	"Store myself on the given stream as a monophonic sound compressed with the given type of compression. The sampling rate is reduced to 22050 samples/second if it is higher."

	| fmt inBufSize samplesPerFrame codec inBuf compressed outSamplingRate audioWriter samplesRemaining outBuf counts byteCount |
	self pause; reset.  "stop playing and return to beginning"

	fmt _ SunAudioFileWriter formatCodeForCompressionType: compressionName.
	inBufSize _ 64000.
	samplesPerFrame _ 1.
	codec _ SunAudioFileWriter codecForFormatCode: fmt.
	codec ifNotNil: [
		samplesPerFrame _ codec samplesPerFrame.
		inBufSize _ inBufSize roundUpTo: (2 * samplesPerFrame).
		compressed _ ByteArray new:
			(inBufSize // samplesPerFrame) * codec bytesPerEncodedFrame].
	inBuf _ SoundBuffer newMonoSampleCount: inBufSize.
	outSamplingRate _ streamSamplingRate.
	streamSamplingRate > 22050 ifTrue: [
		streamSamplingRate = 44100 ifFalse: [self error: 'unexpected MP3 sampling rate'].
		outSamplingRate _ 22050].

	"write audio header"
	audioWriter _ SunAudioFileWriter onStream: aBinaryStream.
	audioWriter writeHeaderSamplingRate: outSamplingRate format: fmt.

	"convert and write sound data"
	'Storing audio...' displayProgressAt: Sensor cursorPoint
		from: 0 to: totalSamples during: [:bar |
			samplesRemaining _ totalSamples.
			[samplesRemaining > 0] whileTrue: [
				bar value: totalSamples - samplesRemaining.
				samplesRemaining < inBuf monoSampleCount ifTrue: [
					inBuf _ SoundBuffer newMonoSampleCount:
						(samplesRemaining roundUpTo: 2 * samplesPerFrame)].
				mpegFile audioReadBuffer: inBuf stream: 0 channel: 0.
				outSamplingRate < streamSamplingRate
					ifTrue: [outBuf _ inBuf downSampledLowPassFiltering: true]
					ifFalse: [outBuf _ inBuf].
				codec
					ifNil: [audioWriter appendSamples: outBuf]
					ifNotNil: [
						counts _ codec
							encodeFrames: (outBuf size // samplesPerFrame)
							from: outBuf at: 1
							into: compressed at: 1.
						byteCount _ counts last.
						byteCount = compressed size
							ifTrue: [audioWriter appendBytes: compressed]
							ifFalse: [audioWriter appendBytes: (compressed copyFrom: 1 to: byteCount)]].
				samplesRemaining _ samplesRemaining - inBuf monoSampleCount]].

	"update audio header"
	audioWriter updateHeaderDataSize.
! !

!StreamingMP3Sound methodsFor: 'accessing' stamp: 'jm 11/16/2001 15:34'!
streamSamplingRate
	"Answer the sampling rate of the MP3 stream."

	^ streamSamplingRate
! !

!StreamingMP3Sound methodsFor: 'accessing' stamp: 'jm 9/26/2000 07:49'!
volume
	"Answer my volume."

	^ volume
! !

!StreamingMP3Sound methodsFor: 'accessing' stamp: 'jm 5/30/2001 16:53'!
volume: aNumber
	"Set my volume to the given number between 0.0 and 1.0."

	volume _ aNumber.
	self createMixer.
! !

AlignmentMorph subclass: #MPEGMoviePlayerMorph
	instanceVariableNames: 'moviePlayer positionSlider volumeSlider'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Player'!

!MPEGMoviePlayerMorph commentStamp: '<historical>' prior: 0!
I provide the user-interface for playing MPEG movies, including play/stop/rewind buttons and volume and position sliders. To create an instance of me, evaluate:

  MPEGMoviePlayerMorph new openInWorld

Then use the "open" button to open an MPEG movie file. This class supplies the front panel; the real work is done by MPEGDisplayMorph and StreamingMP3Sound.

!

!MPEGMoviePlayerMorph class methodsFor: 'scripting' stamp: 'yo 1/14/2005 19:17'!
additionsToViewerCategories
	"Answer a list of (<categoryName> <list of category specs>) pairs that characterize the phrases this kind of morph wishes to add to various Viewer categories."

	^ #(
	(basic
		(
			(command play 'Start playing the movie/sound')
			(command stop 'Stop playing the movie/sound')
			(command rewind 'Rewind the movie/sound')))

	(#'movie controls'
		(
			(slot videoFileName	'The name for the video file' 											String	readWrite Player getVideoFileName Player setVideoFileName:)
			(slot subtitlesFileName	'The name for the subtitles file' 										String	readWrite Player getSubtitlesFileName Player setSubtitlesFileName:)
			(slot position 			'A number representing the current position of the movie/sound.'	Number readWrite Player getPosition Player setPosition:)
			(slot volume 			'A number representing the volume of the movie.' 					Number readWrite Player getVolume Player setVolume:)

			(command play 'Start playing the movie/sound')
			(command playUntilPosition: 'Play until the given position, then stop' Number)			
			(command stop 'Stop playing the movie/sound')
			(command rewind 'Rewind the movie/sound')

			(slot isRunning 'Whether the movie/sound is being played' Boolean readOnly	Player getIsRunning unused unused)
			(slot repeat 'Whether the movie/sound will play in an endless loop' Boolean readWrite	Player getRepeat Player setRepeat:)
			(slot totalFrames 'Length of this movie in number of frames' Number readOnly	Player getTotalFrames unused unused)
			(slot totalSeconds 'Length of this movie in seconds' Number readOnly	Player getTotalSeconds unused unused)
			(slot frameGraphic 'A graphic for the current frame' Graphic readOnly Player getFrameGraphic unused unused)
		)
	)

)! !

!MPEGMoviePlayerMorph class methodsFor: 'parts bin' stamp: 'jm 12/17/2001 14:58'!
descriptionForPartsBin
	^ self partName:	'MPEGPlayer'
		categories:		#('Multimedia')
		documentation:	'A player for MPEG and JPEG movies '! !

!MPEGMoviePlayerMorph class methodsFor: 'fileIn/Out' stamp: 'nk 7/16/2003 15:54'!
fileReaderServicesForFile: fullName suffix: suffix 

	^((MPEGPlayer registeredVideoFileSuffixes includes: suffix )
		or: [ (MPEGPlayer registeredAudioFileSuffixes includes: suffix)
			or: [ suffix = '*' ]] )
		ifTrue: [ self services ]
		ifFalse: [ #() ]! !

!MPEGMoviePlayerMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 10:40'!
initialize 
    "MPEGMoviePlayerMorph initialize."

    FileList registerFileReader: self.

	self registerInFlapsRegistry.	! !

!MPEGMoviePlayerMorph class methodsFor: 'registering' stamp: 'sw 9/7/2004 18:22'!
openFile: aFileName
	"Open the given file (if not nil) in an instance of the receiver."

     | wrapper |
	aFileName ifNil: [^ Beeper beep].
     wrapper _ self openOn: aFileName.
 	wrapper openInWorld.
     ^ wrapper! !

!MPEGMoviePlayerMorph class methodsFor: 'registering' stamp: 'sw 9/7/2004 18:22'!
playFile: aFileName
	"Play the given file (if not nil) in an MPEGMoviePlayerMorph"

     | wrapper |
	aFileName ifNil: [^ Beeper beep].
     wrapper _ self openOn: aFileName.
     wrapper moviePlayer startPlaying. 
     "wrapper openInWindow."
	wrapper openInWorld.
     ^wrapper! !

!MPEGMoviePlayerMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 19:17'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: #(MPEGMoviePlayerMorph	authoringPrototype		'Movie Player'		'A Player for MPEG movies') 
						forFlapNamed: 'Widgets']
! !

!MPEGMoviePlayerMorph class methodsFor: 'registering' stamp: 'sw 9/7/2004 18:17'!
serviceOpenInMPEGPlayer
	"Answer a service for opening a file in an MPEGMoviePlayer"

	^ SimpleServiceEntry 
		provider: self 
		label: 'open'
		selector: #openFile:
		description: 'open file in an MPEG player'
		buttonLabel: 'open'! !

!MPEGMoviePlayerMorph class methodsFor: 'registering' stamp: 'bkv 11/21/2002 11:47'!
servicePlayInMPEGPlayer
	"Answer a service for opening in a MPEG player"

	^ SimpleServiceEntry
		provider: self
		label: 'play in MPEG player'
		selector: #playFile: 
		description: 'play in MPEG player'
		buttonLabel: 'play'! !

!MPEGMoviePlayerMorph class methodsFor: 'fileIn/Out' stamp: 'sw 9/7/2004 18:17'!
services 
	"Answer the receiver's services"

     ^ OrderedCollection with: self servicePlayInMPEGPlayer with: self serviceOpenInMPEGPlayer
! !

!MPEGMoviePlayerMorph class methodsFor: 'class initialization' stamp: 'ads 7/30/2003 16:07'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl | cl unregisterQuadsWithReceiver: self].

	FileList unregisterFileReader: self.! !

!MPEGMoviePlayerMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 20:40'!
addButtonRow
	"private - add the button row"

	| r |
	r _ AlignmentMorph newRow vResizing: #shrinkWrap; color: Color transparent; listCentering: #center.
	r addMorphBack: (self buttonName: 'Menu' translated action: #invokeMenu).
	r addMorphBack: (Morph new extent: 3@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Open' translated action: #openMPEGFile).
	r addMorphBack: (Morph new extent: 3@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Rewind' translated action: #rewindMovie).
	r addMorphBack: (Morph new extent: 3@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Play' translated action: #startPlaying).
	r addMorphBack: (Morph new extent: 3@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Stop' translated action: #stopPlaying).
	r addMorphBack: (Morph new extent: 3@1; color: Color transparent).
"
	r addMorphBack: (self buttonName: '<' action: #previousFrame).
	r addMorphBack: (Morph new extent: 3@1; color: Color transparent).
	r addMorphBack: (self buttonName: '>' action: #nextFrame).
	r addMorphBack: (Morph new extent: 3@1; color: Color transparent).
	r addMorphBack: (self buttonName: 'Subtitles' translated action: #openSubtitlesFile).
	r addMorphBack: (Morph new extent: 3@1; color: Color transparent).
"
	r addMorphBack: (self buildQuitButton).

	self addMorphBack: r.
! !

!MPEGMoviePlayerMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 20:40'!
addPositionSlider
	"private - add the position slider"

	| r |
	positionSlider _ SimpleSliderMorph new
		color: (Color r: 0.71 g: 0.871 b: 1.0);
		extent: 200@2;
		target: moviePlayer;
		actionSelector: #moviePosition:;
		adjustToValue: 0.
	r _ AlignmentMorph newRow
		color: Color transparent;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #leftCenter; listCentering: #center;
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: 'start ' translated).
	r addMorphBack: positionSlider.
	r addMorphBack: (StringMorph contents: ' end' translated).
	self addMorphBack: r.
! !

!MPEGMoviePlayerMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 20:40'!
addVolumeSlider
	"private - add the volume slider"

	| r |
	volumeSlider _ SimpleSliderMorph new
		color: (Color r: 0.71 g: 0.871 b: 1.0);
		extent: 200@2;
		target: moviePlayer;
		actionSelector: #volume:;
		adjustToValue: 0.5.
	r _ AlignmentMorph newRow
		color: Color transparent;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #leftCenter; listCentering: #center;
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: '  soft ' translated).
	r addMorphBack: volumeSlider.
	r addMorphBack: (StringMorph contents: ' loud' translated).
	self addMorphBack: r.
! !

!MPEGMoviePlayerMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 20:41'!
buildQuitButton
	"private - create the [quit] button"
	^ self
		buttonName: 'Quit' translated
		target: self
		action: #quit! !

!MPEGMoviePlayerMorph methodsFor: 'private' stamp: 'jm 6/3/2001 23:41'!
buttonFillStyle

	| fill |
	fill _ GradientFillStyle ramp: {
		0.0->(Color r: 0.742 g: 0.903 b: 1.0).
		1.0->(Color r: 0.516 g: 0.71 b: 1.0)
	}.
	fill origin: self bounds topLeft + (14@3).
	fill direction: 8@6.
	fill radial: false.
	^ fill
! !

!MPEGMoviePlayerMorph methodsFor: 'private' stamp: 'dgd 3/7/2004 19:22'!
buttonName: aString action: aSymbol

	^ self buttonName: aString target: moviePlayer action: aSymbol
! !

!MPEGMoviePlayerMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 20:41'!
buttonName: aString target: anObject action: selector
	"private - create a button"

	^ SimpleButtonMorph new
		target: anObject;
		label: aString;
		actionSelector: selector;
		color: (Color gray: 0.8);  "old color"
		fillStyle: self buttonFillStyle;
		borderWidth: 0;
		borderColor: #raised.
! !

!MPEGMoviePlayerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:39'!
defaultColor
	"answer the default color/fill style for the receiver"
		| fill |
	fill _ GradientFillStyle ramp: {0.0
					-> (Color
							r: 0.355
							g: 0.548
							b: 1.0). 1.0
					-> (Color
							r: 0.774
							g: 0.935
							b: 1.0)}.
	fill origin: self bounds topLeft + (61 @ 7).
	fill direction: 33 @ 37.
	fill radial: false.
	^ fill! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 2/15/2004 22:11'!
defaultFloatPrecisionFor: aGetSelector 
	"Answer a number indicating the default float precision to be  
	used in a numeric readout for which the receiver provides the 
	data. Individual morphs can override this. Showing fractional  
	values for readouts of getCursor was in response to an explicit 
	request from ack"
	aGetSelector == #getVolume
		ifTrue: [^ 0.01].
	aGetSelector == #getPosition 
		ifTrue: [^ 0.001].
	^ super defaultFloatPrecisionFor: aGetSelector! !

!MPEGMoviePlayerMorph methodsFor: 'submorphs-add/remove' stamp: 'dgd 3/8/2004 20:40'!
delete
	"the receiver is being deleted"
	moviePlayer stopPlaying.
	moviePlayer closeFile.
	""
	super delete! !

!MPEGMoviePlayerMorph methodsFor: 'drawing' stamp: 'jm 11/13/2001 09:12'!
drawOn: aCanvas
	"Optimization: Do not draw myself if the movie player is one of my submorphs and the only damage is contained within it. This avoids overdrawing while playing a movie."

	((moviePlayer owner == self) and:
	 [moviePlayer bounds containsRect: aCanvas clipRect])
		ifFalse: [super drawOn: aCanvas].
! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'jm 9/28/2004 17:06'!
getCurrentFrameForm
	"Answer a Form containing the current frame scaled to the size of my display."

	^ moviePlayer currentFrameScaled
! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'jm 9/28/2004 16:59'!
getCurrentFrameImageMorph
	"Answer an ImageMorph containing the current frame scaled to the size of my display."

	^ ImageMorph new image: (moviePlayer currentFrameScaled)
! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 2/15/2004 22:03'!
getIsRunning
	"answer whateve the receiver is running"
	^ moviePlayer isRunning! !

!MPEGMoviePlayerMorph methodsFor: 'accessing' stamp: 'dgd 2/15/2004 21:11'!
getPosition
	"answer the receiver's movie position"
	^ positionSlider getScaledValue! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 2/15/2004 22:06'!
getRepeat
	"answer whateve the receiver is running"
	^ moviePlayer repeat! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 3/8/2004 21:42'!
getSubtitlesFileName
	"answer the receiver's subtitlesFileName"
	^ moviePlayer subtitlesFileShortName! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 3/8/2004 21:41'!
getVideoFileName
	"answer the receiver's videoFileName"
	^ moviePlayer videoFileShortName! !

!MPEGMoviePlayerMorph methodsFor: 'accessing' stamp: 'dgd 2/15/2004 21:53'!
getVolume
	"answer the receiver's movie position"
	^ self volumeSlider isNil ifFalse:[self volumeSlider getScaledValue] ifTrue:[0.0]! !

!MPEGMoviePlayerMorph methodsFor: 'accessing' stamp: 'dgd 2/15/2004 20:23'!
guessVolumeSlider
	"private - look for a morph that is the receiver's volumeSlider"
	^ self allMorphs
		detect: [:each | "first look in my own morphs"
			each class == SimpleSliderMorph
				and: [each actionSelector == #volume:]]
		ifNone: [| w | 
			"second try, look all over the world (if any)"
			w := self world.
			w isNil
				ifFalse: [""
					w allMorphs
						detect: [:each | ""
							each class == SimpleSliderMorph
								and: [each actionSelector == #volume:]
								and: [each target == moviePlayer]]
						ifNone: []]]! !

!MPEGMoviePlayerMorph methodsFor: 'event handling' stamp: 'dgd 3/7/2004 22:34'!
handlesKeyboard: evt
	^ moviePlayer handlesKeyboard: evt! !

!MPEGMoviePlayerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:40'!
initialize
"initialize the state of the receiver"
	super initialize.
	""
	self hResizing: #shrinkWrap;
		 vResizing: #shrinkWrap.
	borderWidth _ 2.
	self listDirection: #topToBottom.
	self cornerStyle: #rounded.
	self layoutInset: 4.
	moviePlayer _ MPEGDisplayMorph new.
	self addMorphFront: moviePlayer.
	self addButtonRow.
	self addVolumeSlider.
	self addPositionSlider.
	self extent: 10 @ 10! !

!MPEGMoviePlayerMorph methodsFor: 'event handling' stamp: 'dgd 3/7/2004 22:34'!
keyStroke: evt 
	moviePlayer keyStroke: evt ! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 2/15/2004 21:30'!
play
	"play the receiver"
	moviePlayer startPlaying! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'jm 10/12/2004 11:14'!
playUntilPosition: finalPosition
	"Play the movie until the given position, then stop."

	moviePlayer playUntilPosition: finalPosition! !

!MPEGMoviePlayerMorph methodsFor: 'private' stamp: 'dgd 3/8/2004 20:39'!
quit
	"quit the receiver"
	self delete! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 2/15/2004 21:31'!
rewind
	"rewind the receiver"
	moviePlayer rewindMovie! !

!MPEGMoviePlayerMorph methodsFor: 'accessing' stamp: 'dgd 2/15/2004 21:11'!
setPosition: aNumber 
	"changes the receiver's movie position"
	| newPosition |
	newPosition := aNumber asFloat min: 1.0 max: 0.0.
	positionSlider value: newPosition.
	moviePlayer moviePosition: newPosition! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 2/15/2004 22:07'!
setRepeat: aBoolean
	"answer whateve the receiver is running"
	moviePlayer repeat: aBoolean! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 3/8/2004 21:42'!
setSubtitlesFileName: aString 
	"change the subtitlesFileName"
	moviePlayer subtitlesFileShortName: aString! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 3/8/2004 21:41'!
setVideoFileName: aString 
	"change the videoFileName"
	moviePlayer videoFileShortName: aString! !

!MPEGMoviePlayerMorph methodsFor: 'accessing' stamp: 'dgd 2/15/2004 21:57'!
setVolume: aNumber 
	"changes the receiver's movie position"
	| newVolume |
	newVolume := aNumber asFloat min: 1.0 max: 0.0.
	self volumeSlider isNil ifFalse:[self volumeSlider value: newVolume].
	moviePlayer volume: newVolume! !

!MPEGMoviePlayerMorph methodsFor: 'stepping' stamp: 'jm 4/6/2001 07:49'!
step
	"Update the position slider from the current movie position."

	positionSlider adjustToValue: moviePlayer moviePosition.
! !

!MPEGMoviePlayerMorph methodsFor: 'stepping' stamp: 'jm 5/30/2001 23:33'!
stepTime
	"Update the position slider a few times a second."

	^ 500
! !

!MPEGMoviePlayerMorph methodsFor: 'e-toy support' stamp: 'dgd 2/15/2004 21:31'!
stop
	"stop the receiver"
	moviePlayer stopPlaying! !

!MPEGMoviePlayerMorph methodsFor: 'accessing' stamp: 'jm 10/12/2004 11:31'!
totalFrames
	"Answer the total number of frames in this movie."

	^ moviePlayer totalFrames! !

!MPEGMoviePlayerMorph methodsFor: 'accessing' stamp: 'jm 10/12/2004 11:31'!
totalSeconds
	"Answer the total number of seconds in this movie."

	^ moviePlayer totalSeconds! !

!MPEGMoviePlayerMorph methodsFor: 'accessing' stamp: 'dgd 2/15/2004 20:13'!
volumeSlider
	"answer the receiver's volumeSlider  
	 
	note: if the instance var is undefined, try to get the sliders  
	from the allMorphs chain. in this way an instance of the  
	receiver created before the instVars was added can works fine"
	^ volumeSlider
		ifNil: [volumeSlider := self guessVolumeSlider]! !

UpdatingTextMorph subclass: #MPEGSubtitlesDisplayer
	instanceVariableNames: 'font'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Player'!

!MPEGSubtitlesDisplayer methodsFor: 'menu' stamp: 'dgd 3/8/2004 20:42'!
changeFont
	"open a dialog to change the receiver's font"
	| newFont |
	newFont := StrikeFont fromUser: self font.
	""
	newFont isNil
		ifFalse: [self font: newFont]! !

!MPEGSubtitlesDisplayer methodsFor: 'menu' stamp: 'dgd 3/8/2004 20:10'!
changeSubtitlesColor
	"offer a ColorPicker to change the subtitles colors"

	ColorPickerMorph new
		choseModalityFromPreference;
		sourceHand: self activeHand;
		target: self;
		selector: #textColor:;
		originalColor: self textColor;
		putUpFor: self currentHand near: self currentHand cursorBounds
! !

!MPEGSubtitlesDisplayer methodsFor: 'target access' stamp: 'dgd 3/8/2004 20:36'!
contentsFromTarget
	"private - answer the contents from the receiver's target"
	| contentsAsText |
	contentsAsText := super contentsFromTarget asText.
	contentsAsText
		addAttribute: (TextFontReference toFont: self font).
	^ contentsAsText! !

!MPEGSubtitlesDisplayer methodsFor: 'accessing' stamp: 'dgd 3/7/2004 21:18'!
font
	"answer the receiver's font"
	^ font
		ifNil: [TextStyle defaultFont] ! !

!MPEGSubtitlesDisplayer methodsFor: 'accessing' stamp: 'dgd 3/8/2004 20:33'!
font: aFont 
	"change the receiver's font"
	font := aFont.
	""
	self contents: ''.
	self contents: self contentsFromTarget! !

!MPEGSubtitlesDisplayer methodsFor: 'initialization' stamp: 'dgd 3/7/2004 21:16'!
initialize
	"initialiaze the receiver"
	super initialize.
	""
font := TextStyle defaultFont.""
	self
		backgroundColor: (Color black alpha: 0.4).
	""
	self margins: 4 @ 2.
	self textColor: Color white.
	self textStyle centered! !

!MPEGSubtitlesDisplayer methodsFor: 'stepping and presenter' stamp: 'dgd 3/7/2004 20:59'!
step
	"update my position"
	super step.
	" 
	if my owner is the mpegplayer, i change my position to  
	bottomCenter"
	self owner == self target
		ifTrue: [| bc | 
			bc := self owner bottomCenter.
			self left: bc x - (self width // 2).
			self bottom: bc y]! !

BlobMorph subclass: #BlobMPEGMorph
	instanceVariableNames: 'mpegLogic form movieDrawArea primary quadNumber'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Player'!

!BlobMPEGMorph commentStamp: '<historical>' prior: 0!
Ok this is a little follow on to David's BlobMorph. Why not embedded a movie in the blob I thought. So with a few minutes of help from John Maloney we have something very interesting.

Enjoy John M McIntosh <johnmci@smalltalkconsulting.com> Dec 2000. (Christmas early)!

!BlobMPEGMorph class methodsFor: 'instance creation' stamp: 'JMM 1/4/2001 11:10'!
buildMorphics: aFileName
	| primary child |

	primary _ (self basicNew quadNumber: 1) initialize.
	self remember: primary.
	primary initializePrimaryMpegPlayer: aFileName.
	primary openInWorld.
	2 to: 4 do: [:i | 
		child _ (self basicNew quadNumber: i) initialize.
		self remember: child.
		child initializeChildMpegPlayer: primary.
		child openInWorld].
	^primary
! !

!BlobMPEGMorph class methodsFor: 'instance creation' stamp: 'JMM 1/4/2001 11:11'!
newWithMovie: aFileName
	| primary |

	primary _ self buildMorphics: aFileName.
	primary playStream: 0.
	^primary

! !

!BlobMPEGMorph class methodsFor: 'instance creation' stamp: 'JMM 1/4/2001 11:12'!
newWithMovieNoSound: aFileName
	| primary |

	primary _ self buildMorphics: aFileName.
	primary playVideoStream: 0.
	^primary

! !

!BlobMPEGMorph methodsFor: 'stepping' stamp: 'JMM 10/19/2000 15:57'!
adjustColors
	^self! !

!BlobMPEGMorph methodsFor: 'drawing' stamp: 'JMM 1/4/2001 11:18'!
drawOn: aCanvas 
	"Display the receiver, a spline curve, approximated by straight line 
	segments. Fill with the MPEG movie"

	| cm f filled quadRect |

	cm _ Bitmap new: 2.
	cm at: 1 put: 0.
	cm at: 2 put: 32767.
	f _ Form extent: self extent depth: 16.
	filled _ self filledForm.
	(BitBlt toForm: f) sourceForm: filled;
		 sourceRect: filled boundingBox;
		
		destRect: (0 @ 0 extent: filled extent);
		 colorMap: cm;
		 combinationRule: Form over;
		 copyBits.
	quadNumber = 1
		ifTrue: [quadRect _ Rectangle origin: form boundingBox topLeft corner: form boundingBox center].
	quadNumber = 2
		ifTrue: [quadRect _ Rectangle origin: form boundingBox topCenter corner: form boundingBox rightCenter].
	quadNumber = 3
		ifTrue: [quadRect _ Rectangle origin: form boundingBox leftCenter corner: form boundingBox bottomCenter].
	quadNumber = 4
		ifTrue: [quadRect _ Rectangle origin: form boundingBox center corner: form boundingBox bottomRight].
	(BitBlt toForm: f) sourceForm: form;
		 sourceRect: quadRect;
		
		destRect: (0 @ 0 extent: f extent);
		 combinationRule: Form and;
		 copyBits.
	aCanvas image: f at: self position.
	self drawBorderOn: aCanvas.
	self drawArrowsOn: aCanvas! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 10/19/2000 16:02'!
form
	^form! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 10/19/2000 16:02'!
form: aForm
	form _ aForm! !

!BlobMPEGMorph methodsFor: 'initialization' stamp: 'aoy 2/15/2003 21:43'!
initialize: primaryFlag mpegPlayer: aMpegPlayerOrFileName 
	| rect sizeToOverLapBoundary |
	primary := primaryFlag.
	rect := self bounds.
	sizeToOverLapBoundary := 3.0.
	mpegLogic := primary 
				ifTrue:  
					[form := Form 
								extent: ((sizeToOverLapBoundary * rect width) 
										@ (sizeToOverLapBoundary * rect height)) truncated
								depth: 32.
					movieDrawArea := SketchMorph withForm: form.
					MPEGPlayer playFile: aMpegPlayerOrFileName onMorph: movieDrawArea]
				ifFalse: 
					[form := aMpegPlayerOrFileName form.
					movieDrawArea := aMpegPlayerOrFileName movieDrawArea.
					aMpegPlayerOrFileName mpegLogic]! !

!BlobMPEGMorph methodsFor: 'initialization' stamp: 'JMM 1/4/2001 11:02'!
initializeBlobShape

	| verts modifier |
	verts _ {59@40. 74@54. 79@74. 77@93. 57@97. 37@97. 22@83. 15@67. 22@50. 33@35. 47@33}.
	modifier _ 0 @ 0.
	(self quadNumber = 2) ifTrue: [ modifier _ 0 @ 75].
	(self quadNumber = 3) ifTrue: [ modifier _ 75 @ 0].
	(self quadNumber = 4) ifTrue: [ modifier _ 75 @ 75].
	verts _ verts + modifier.
	self 
		vertices: verts
		color: self color
		borderWidth: 1
		borderColor: Color black! !

!BlobMPEGMorph methodsFor: 'initialization' stamp: 'JMM 1/4/2001 10:54'!
initializeChildMpegPlayer: aMpegPlayerOrFileName  
	self initialize: false mpegPlayer: aMpegPlayerOrFileName ! !

!BlobMPEGMorph methodsFor: 'initialization' stamp: 'JMM 1/4/2001 10:54'!
initializePrimaryMpegPlayer: aMpegPlayerOrFileName  
	self initialize: true mpegPlayer: aMpegPlayerOrFileName  ! !

!BlobMPEGMorph methodsFor: 'stepping' stamp: 'JMM 12/19/2000 15:39'!
limitRange: verts
	" limit radius to range 20-120; limit interpoint angle to surrounding angles with max of twice of average separation. "
	| cent new prevn nextn prevDeg nextDeg thisDeg dincr |

	cent := self bounds center.
	new := Array new: verts size.
	dincr := 360 // verts size.
	verts doWithIndex: [ :pt :n |

		"Find prev/next points, allowing for wrapping around "
		prevn := n-1 < 1 ifTrue: [new size] ifFalse: [n-1].
		nextn := n+1 > new size ifTrue: [1] ifFalse: [n+1].

		"Get prev/this/next point's angles "
		prevDeg := ((verts at: prevn)-cent) degrees.
		thisDeg := ((verts at: n)-cent) degrees.
		nextDeg := ((verts at: nextn)-cent) degrees.

		"Adjust if this is where angles wrap from 0 to 360"
		(thisDeg - prevDeg) abs > 180 ifTrue: [ prevDeg := prevDeg - 360 ].
		(thisDeg - nextDeg) abs > 180 ifTrue: [ nextDeg := nextDeg + 360 ].

		"Put adjusted point into new collection"
		new at: n put: cent +
			(self selfPolarPointRadius: ((((pt - cent) r) min: 60) max: 20) "was min: 80"
				degrees: (((thisDeg min: nextDeg-5) max: prevDeg+5) min: dincr*2+prevDeg)) ].
	^ new
! !

!BlobMPEGMorph methodsFor: 'stepping' stamp: 'JMM 12/19/2000 15:29'!
mergeBlobs
	^self! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 12/19/2000 16:01'!
movieDrawArea
	^movieDrawArea! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 10/19/2000 15:54'!
mpegLogic
	^mpegLogic! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 10/19/2000 15:54'!
mpegLogic: aValue
	mpegLogic _ aValue! !

!BlobMPEGMorph methodsFor: 'drawing' stamp: 'JMM 1/4/2001 11:07'!
playStream: aStream
	mpegLogic playStream: aStream.
! !

!BlobMPEGMorph methodsFor: 'drawing' stamp: 'JMM 12/19/2000 16:41'!
playVideoStream: aStream
	mpegLogic playVideoStream: aStream.
! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 12/19/2000 15:45'!
primary
	^primary! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 12/19/2000 21:52'!
quadNumber
	^quadNumber! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 1/4/2001 10:53'!
quadNumber: aNumber
	quadNumber _ aNumber! !

!BlobMPEGMorph methodsFor: 'testing' stamp: 'JMM 10/19/2000 16:29'!
stepTime
	^1.0 / (self mpegLogic videoFrameRate: self stream) * 1000! !

!BlobMPEGMorph methodsFor: 'access' stamp: 'JMM 10/19/2000 15:59'!
stream
	^0! !

ImageMorph subclass: #MovieFrameSyncMorph
	instanceVariableNames: 'moviePlayerMorph frameNumber'
	classVariableNames: 'ColorIndex'
	poolDictionaries: ''
	category: 'Movies-Player'!

!MovieFrameSyncMorph commentStamp: '<historical>' prior: 0!
This class and its subclasses act to syncronize movie players with the progress of a MIDI score and further background can be found in the comment for MoviePlayerMorph.

There are two main forms of synchroniztion.

MovieFrameSyncMorphs are used where you have a MIDI score open with a piano roll, and a separate MoviePlayer Morph.  A MovieFrameSyncMorph can be torn off from the MoviePlayer with a shift-drag gesture or menu command, and can then be dropped into a MIDI score.  They are used to start a movie player at a given place in the score, and then to stretch or compress the movie frame rate so that certain frames are synchronized with the corresponding points in the score.

MovieClipStartMorphs (q.v.) and MovieClipEndMorphs are used for the other kind of synchronization, namely where you wish to assemble a number of movie clips in a sequence.!

MovieFrameSyncMorph subclass: #MovieClipStartMorph
	instanceVariableNames: 'movieClipPlayer movieClipFileName soundTrackFileName soundTrackPlayerReady soundTrackMorph soundTrackTimeScale scoreEvent endMorph clipColor colorMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Player'!

!MovieClipStartMorph commentStamp: '<historical>' prior: 0!
This class and its subclasses act to syncronize movie players with the progress of a MIDI score and further background can be found in the comment for MoviePlayerMorph.

MovieClipStartMorphs are used where you have a movie clip player plex.  You can create such a plex by opening a midi score, spawning a piano roll (by the button of that name), and then choosing 'add movie clip player' from the piano roll menu.  Much simpler, you can open a MoviePlayerMorph as a new morph, and then choose 'make a new movie' from its menu.  This plex then serves as a central editable score for composing movie clips.

To add a new movie clip to the score, open a movie from the fileList (or as a new morph followed by 'open a movie').  Play or otherwise position the clip to the desired starting position, and then tear off a thumbnail and drop it into the score at the desired time.  The starting thumbnail (MovieClipStartMorph) will appear in the score, tied to an endMorph by a colored stripe.  The ending time will be chosen based on the total length of the clip, a default starting clip length (200 frames), and possible interference with other clips that follow it.

To reposition a clip, you can pick up its clipStart with halo black handle, and drop it elsewhere.  The rest of the clip will follow as best it can.  To delete a clip, delete its clipStart.  To change the duration of a clip, play the composition up to some point in that clip, and pause it.  Then use the  controls on the central movie player to move forward or backward to the desired ending frame, and choose 'end clip here' from the player menu.!

MovieClipStartMorph subclass: #MovieClipEndMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Player'!

!MovieClipEndMorph commentStamp: '<historical>' prior: 0!
The idea is that soon we will show the soundtrack extending between the start cue and the end morph.!

!MovieClipEndMorph methodsFor: 'piano rolls' stamp: 'di 10/26/2000 00:25'!
addMorphsTo: morphList pianoRoll: pianoRoll eventTime: t betweenTime: leftTime and: rightTime

	"Ignored -- all display is done by the starting morph -- see superclass"! !

!MovieClipEndMorph methodsFor: 'piano rolls' stamp: 'di 10/11/2000 23:06'!
encounteredAtTime: ticks inScorePlayer: scorePlayer atIndex: index inEventTrack: track secsPerTick: secsPerTick

	movieClipPlayer ifNotNil:
		["If being shown as a clip, then tell the clipPlayer to stop showing this clip"
		movieClipPlayer stopRunning]! !

!MovieClipEndMorph methodsFor: 'as yet unclassified' stamp: 'di 10/22/2000 17:28'!
movieFileName: movieFileName image: aForm player: aMoviePlayer frameNumber: n

	movieClipFileName _ movieFileName.
	self image: aForm frameNumber: n.
	moviePlayerMorph _ movieClipPlayer _ aMoviePlayer.
	scoreEvent _ AmbientEvent new morph: self.
! !

!MovieClipEndMorph methodsFor: 'piano rolls' stamp: 'di 10/22/2000 12:43'!
pauseFrom: scorePlayer

	"Ignored"! !

!MovieClipEndMorph methodsFor: 'piano rolls' stamp: 'di 10/22/2000 12:43'!
resetFrom: scorePlayer

	"Ignored"! !

!MovieClipEndMorph methodsFor: 'piano rolls' stamp: 'di 10/22/2000 12:43'!
resumeFrom: scorePlayer

	"Ignored"! !

!MovieClipEndMorph methodsFor: 'drawing' stamp: 'di 10/23/2000 10:42'!
thumbnailHeight

	^ 30! !

!MovieClipStartMorph methodsFor: 'piano rolls' stamp: 'dgd 2/22/2003 14:09'!
addMorphsTo: morphList pianoRoll: pianoRoll eventTime: t betweenTime: leftTime and: rightTime 
	"This code handles both the start and end morphs."

	| startX endX h delta |
	self startTime > rightTime 
		ifTrue: [^self	"Start time has not come into view."].
	self endTime < leftTime ifTrue: [^self	"End time has passed out of view."].
	startX := pianoRoll xForTime: self startTime.
	endX := pianoRoll xForTime: self endTime.
	h := self colorMargin.	"Height of highlight bar over thumbnails."
	morphList add: (self align: self bottomLeft
				with: startX @ (pianoRoll bottom - pianoRoll borderWidth - h)).
	morphList 
		add: (endMorph align: endMorph bounds rightCenter with: endX @ self center y).
	morphList add: (self colorMorph 
				bounds: (self topLeft - (0 @ h) corner: endMorph right @ (self bottom + h))).
	(soundTrackMorph isNil and: [moviePlayerMorph scorePlayer isNil]) 
		ifFalse: 
			["Wants a sound track"

			(soundTrackMorph isNil or: [pianoRoll timeScale ~= soundTrackTimeScale]) 
				ifTrue: 
					["Needs a new sound track"

					self buildSoundTrackMorphFor: pianoRoll].
			morphList add: (soundTrackMorph align: soundTrackMorph bottomLeft
						with: colorMorph topLeft).
			self soundTrackOnBottom 
				ifTrue: 
					[soundTrackMorph align: soundTrackMorph bottomLeft with: self bottomLeft.
					delta := 0 @ self soundTrackHeight.
					self position: self position - delta.
					endMorph position: endMorph position - delta.
					colorMorph position: colorMorph position - delta]]! !

!MovieClipStartMorph methodsFor: 'private' stamp: 'di 10/21/2000 23:28'!
buildSoundTrackMorphFor: pianoRoll
	| stopTime soundTrackForm startTime samplesPerTick samplesPerMs |
	soundTrackTimeScale _ pianoRoll timeScale.  "pixels per tick"
	samplesPerTick _ moviePlayerMorph scorePlayer originalSamplingRate   "Samples per sec"
						* pianoRoll scorePlayer secsPerTick.  "secs per tick"
	samplesPerMs _ moviePlayerMorph scorePlayer originalSamplingRate / 1000.0.
	startTime _ frameNumber * moviePlayerMorph msPerFrame.  "ms"
	stopTime _ endMorph frameNumber * moviePlayerMorph msPerFrame.
	soundTrackForm _ moviePlayerMorph scorePlayer
		volumeForm: self soundTrackHeight
		from: (startTime * samplesPerMs) rounded
		to: (stopTime * samplesPerMs) rounded
		nSamplesPerPixel: samplesPerTick / soundTrackTimeScale.
	^ soundTrackMorph _ ImageMorph new image: soundTrackForm! !

!MovieClipStartMorph methodsFor: 'drawing' stamp: 'di 10/23/2000 10:46'!
colorMargin
	"How far the clip span color highlight extends outside the starting clip"

	^ 5! !

!MovieClipStartMorph methodsFor: 'drawing' stamp: 'di 10/23/2000 10:14'!
colorMorph
	colorMorph ifNotNil: [^ colorMorph].

	"Make up a morph to highlight the span of this clip."
	ColorIndex _ (ColorIndex ifNil: [0]) + 2 \\ 8 + 1.
	^ colorMorph _ Morph newBounds: (0@0 extent: 9@9) color: ((Color wheel: 8) at: ColorIndex)
! !

!MovieClipStartMorph methodsFor: 'submorphs-add/remove' stamp: 'di 10/23/2000 10:24'!
delete
	(owner isKindOf: PianoRollScoreMorph) ifTrue:
		[owner score removeAmbientEventWithMorph: self.
		endMorph ifNotNil: [owner score removeAmbientEventWithMorph: endMorph]].
	endMorph ifNotNil: [endMorph delete].
	soundTrackMorph ifNotNil: [soundTrackMorph delete].
	colorMorph ifNotNil: [colorMorph delete].
	super delete.
! !

!MovieClipStartMorph methodsFor: 'piano rolls' stamp: 'di 10/26/2000 00:09'!
encounteredAtTime: ticks inScorePlayer: scorePlayer atIndex: index inEventTrack: track secsPerTick: secsPerTick

	"If being shown as a clip, then tell the clipPlayer to start showing this clip"
	movieClipPlayer setCueMorph: self.
	movieClipPlayer openFileNamed: movieClipFileName
			withScorePlayer: soundTrackPlayerReady copy
			andPlayFrom: frameNumber.
! !

!MovieClipStartMorph methodsFor: 'access' stamp: 'di 10/19/2000 14:27'!
endMorph

	^ endMorph! !

!MovieClipStartMorph methodsFor: 'events' stamp: 'di 10/22/2000 20:24'!
endTime

	^ endMorph scoreEvent time! !

!MovieClipStartMorph methodsFor: 'events' stamp: 'di 10/22/2000 20:27'!
frameAtTick: time
	"Return the frame number corresponding to the given tick time"

	^ frameNumber +
		((time - self startTime) asFloat
			/ (self endTime - self startTime)
			* (endMorph frameNumber - frameNumber)) asInteger! !

!MovieClipStartMorph methodsFor: 'access' stamp: 'di 10/19/2000 16:41'!
frameNumber: newFrame

	frameNumber _ newFrame! !

!MovieClipStartMorph methodsFor: 'initialization' stamp: 'di 10/23/2000 10:40'!
image: aForm frameNumber: n

	self image: (aForm magnifyBy: self thumbnailHeight asFloat / aForm height).
	frameNumber _ n.! !

!MovieClipStartMorph methodsFor: 'dropping/grabbing' stamp: 'di 10/29/2000 09:23'!
justDroppedInto: newOwner event: evt
	| pianoRoll syncMorph |
	"When dropping this morph into a pianoRoll, add a corresponding
	event to the score so that it will always appear when played,
	in addition to possibly triggering other actions"

	(newOwner isKindOf: PianoRollScoreMorph)
	ifTrue:
		[pianoRoll _ newOwner.
		pianoRoll movieClipPlayer ifNil:
			["This PianoRoll is not a clip player -- replace me by a SyncMorph"
			syncMorph _ MovieFrameSyncMorph new
						image: image
						player: moviePlayerMorph
						frameNumber: frameNumber.
			pianoRoll replaceSubmorph: self by: syncMorph.
			"rewrite to use justDroppedInto:..."
			pianoRoll score removeAmbientEventWithMorph: self;
					addAmbientEvent: (scoreEvent
						morph: syncMorph;
						time: (pianoRoll timeForX: self left)).
			^ self].

		self movieClipPlayer: pianoRoll movieClipPlayer.
		self setTimeInScore: pianoRoll score
					near: (pianoRoll timeForX: self left).
		self endTime > newOwner scorePlayer durationInTicks ifTrue:
			[newOwner scorePlayer updateDuration]]
	ifFalse:
		["Dropped it somewhere else -- delete related morphs"
		endMorph ifNotNil: [endMorph delete].
		soundTrackMorph ifNotNil: [soundTrackMorph delete]].

	super justDroppedInto: newOwner event: evt
! !

!MovieClipStartMorph methodsFor: 'access' stamp: 'di 10/4/2000 17:18'!
movieClipPlayer: aMoviePlayerMorph
	"For now, these morphs work both as a sync point in a long movie, and
	as a cue point for a short clip in a longer score.
	To a cue point, this method provides a reference to the clip player."

	movieClipPlayer _ aMoviePlayerMorph! !

!MovieClipStartMorph methodsFor: 'initialization' stamp: 'di 10/23/2000 10:32'!
movieFileName: movieFileName soundTrackFileName: soundFileName
			image: aForm player: aMoviePlayer frameNumber: n
	movieClipFileName _ movieFileName.
	soundTrackFileName _ soundFileName.
	self image: aForm frameNumber: n.
	moviePlayerMorph _ aMoviePlayer.
	soundTrackPlayerReady _ moviePlayerMorph scorePlayer copy.
	scoreEvent _ AmbientEvent new morph: self.
! !

!MovieClipStartMorph methodsFor: 'access' stamp: 'di 10/25/2000 22:02'!
relatedPlayer
 
	^ movieClipPlayer! !

!MovieClipStartMorph methodsFor: 'piano rolls' stamp: 'dgd 2/22/2003 14:09'!
resetFrom: scorePlayer 
	(movieClipPlayer cueMorph isNil 
		or: [self startTime < movieClipPlayer cueMorph startTime]) 
			ifTrue: 
				[movieClipPlayer
					openFileNamed: movieClipFileName
						withScorePlayer: soundTrackPlayerReady copy
						andPlayFrom: frameNumber;
					setCueMorph: self;
					step;
					pauseFrom: scorePlayer]! !

!MovieClipStartMorph methodsFor: 'piano rolls' stamp: 'di 10/25/2000 22:03'!
resumeFrom: scorePlayer

	| time |
	"New movie clip style of use."
	time _ scorePlayer ticksSinceStart.
	time < self startTime ifTrue: [^ self].  "It's not my time yet"
	time > self endTime ifTrue: [^ self].  "It's past my time"

	"The player is starting in the midst of this clip."
	movieClipPlayer openFileNamed: movieClipFileName
				withScorePlayer: soundTrackPlayerReady copy
				andPlayFrom: (self frameAtTick: time);
		setCueMorph: self.
! !

!MovieClipStartMorph methodsFor: 'access' stamp: 'di 10/19/2000 12:24'!
scoreEvent
	^ scoreEvent! !

!MovieClipStartMorph methodsFor: 'private' stamp: 'di 10/29/2000 08:02'!
setEndFrameNumber: frame

	self setEndFrameNumber: frame tickTime: nil! !

!MovieClipStartMorph methodsFor: 'private' stamp: 'di 10/29/2000 09:17'!
setEndFrameNumber: frameOrNil tickTime: timeOrNil
	"May be called with either time or frame being nil,
	in which case the other will br computed."

	| pianoRoll frame time |
	pianoRoll _ movieClipPlayer pianoRoll.
	frame _ frameOrNil ifNil:
		[frameNumber + 
			((timeOrNil - self startTime)
			* (pianoRoll scorePlayer secsPerTick*1000.0)
			/ moviePlayerMorph msPerFrame) asInteger - 1].
	time _ timeOrNil ifNil:
		[self startTime +   "in ticks"
			(pianoRoll scorePlayer ticksForMSecs:
			(frameOrNil - frameNumber) * moviePlayerMorph msPerFrame)].
	endMorph ifNil:
		[endMorph _ MovieClipEndMorph new
			movieFileName: movieClipFileName
			image: (moviePlayerMorph pageFormForFrame: frame)
			player: movieClipPlayer
			frameNumber: frame]
		ifNotNil:
		[endMorph image: (moviePlayerMorph pageFormForFrame: frame)
			frameNumber: frame].

	endMorph scoreEvent time: time.
	pianoRoll score removeAmbientEventWithMorph: endMorph;
		addAmbientEvent: endMorph scoreEvent.
	soundTrackMorph _ nil.  "Force it to be recomputed."
	pianoRoll rebuildFromScore
! !

!MovieClipStartMorph methodsFor: 'private' stamp: 'di 10/30/2000 08:21'!
setTimeInScore: score near: dropTime
	"Find a time to place this clip that does not overlap other clips.
	So, if I start in the middle of another clip, move me to the end of it,
	and if I start very soon after another clip, put me right at the end.
	Then, if my end goes beyond the start of another clip, shorten me
	so I end right before that clip."

	| startTime endTime delta endFrame |
	startTime _ dropTime.
	endMorph ifNil: [endFrame _ moviePlayerMorph frameCount]
			ifNotNil: [endFrame _ endMorph frameNumber].
	endTime _ startTime   "in ticks"
		+ (movieClipPlayer pianoRoll scorePlayer ticksForMSecs:
			(endFrame - frameNumber)
			* moviePlayerMorph msPerFrame).
	score eventMorphsDo:
		[:m | (m ~~ self and: [m isMemberOf: self class]) ifTrue:
				[((startTime between: m startTime and: m endTime)
					or: [startTime between: m endTime and: m endTime+50])
					ifTrue: ["If I start in the middle of another clip, or a little
							past its end, move me exactly to the end of it"
							delta _ (m endTime + 1) - startTime.
							startTime _ startTime + delta.
							endTime _ endTime + delta].
				(endTime between: m startTime and: m endTime)
					ifTrue: ["If my end goes overlaps another clip, shorten me so I fit."
							endTime _ m startTime - 1].
				]].
	scoreEvent time: startTime.
	score removeAmbientEventWithMorph: self;
			addAmbientEvent: scoreEvent.
	self setEndFrameNumber: endFrame tickTime: endTime.
! !

!MovieClipStartMorph methodsFor: 'drawing' stamp: 'di 10/23/2000 10:23'!
soundTrackHeight

	^ 40! !

!MovieClipStartMorph methodsFor: 'drawing' stamp: 'di 10/23/2000 10:58'!
soundTrackOnBottom  "a local preference during test"

	^ false! !

!MovieClipStartMorph methodsFor: 'events' stamp: 'di 10/22/2000 20:24'!
startTime

	^ scoreEvent time! !

!MovieClipStartMorph methodsFor: 'drawing' stamp: 'di 10/23/2000 12:32'!
thumbnailHeight

	^ 60! !

!MovieFrameSyncMorph methodsFor: 'piano rolls' stamp: 'di 10/25/2000 22:36'!
addMorphsTo: morphList pianoRoll: pianoRoll eventTime: t betweenTime: leftTime and: rightTime

	| leftX |
	t > rightTime ifTrue: [^ self  "Start time has not come into view."].  
	leftX _ pianoRoll xForTime: t.
	(leftX + self width) < pianoRoll left ifTrue: [^ self  "End time has passed out of view."].
	morphList add: 
		(self align: self bottomLeft
			with: leftX @ (pianoRoll bottom - pianoRoll borderWidth)).
! !

!MovieFrameSyncMorph methodsFor: 'piano rolls' stamp: 'dgd 2/21/2003 22:58'!
encounteredAtTime: ticks inScorePlayer: scorePlayer atIndex: index inEventTrack: track secsPerTick: secsPerTick 
	"Set frame number and milliseconds since start in case of drift"

	| next |
	moviePlayerMorph frameNumber: frameNumber
		msSinceStart: scorePlayer millisecondsSinceStart.

	"If there is a later sync point, set the appropriate frame rate until then."
	(next := self nextSyncEventAfter: index inTrack: track) isNil 
		ifFalse: 
			[moviePlayerMorph msPerFrame: (next time - ticks) * secsPerTick * 1000.0 
						/ (next morph frameNumber - self frameNumber)]! !

!MovieFrameSyncMorph methodsFor: 'access' stamp: 'di 8/6/1998 14:13'!
frameNumber
	^ frameNumber! !

!MovieFrameSyncMorph methodsFor: 'initialization' stamp: 'di 8/5/1998 17:04'!
image: aForm player: aMoviePlayer frameNumber: n
	self image: aForm.
	moviePlayerMorph _ aMoviePlayer.
	frameNumber _ n! !

!MovieFrameSyncMorph methodsFor: 'dropping/grabbing' stamp: 'di 10/29/2000 09:22'!
justDroppedInto: newOwner event: evt 
	| pianoRoll |
	"When dropping this morph into a pianoRoll, add a corresponding
	event to the score so that it will always appear when played,
	in addition to possibly triggering other actions"

	(self isMemberOf: MovieFrameSyncMorph) ifFalse:
		[^ super justDroppedInto: newOwner event: evt].

	(newOwner isKindOf: PianoRollScoreMorph)
	ifTrue:
		["Legacy code for existing sync morphs"
		pianoRoll _ newOwner.
		pianoRoll score
			removeAmbientEventWithMorph: self;
			addAmbientEvent: (AmbientEvent new
						morph: self;
						time: (pianoRoll timeForX: self left))].

	super justDroppedInto: newOwner event: evt
! !

!MovieFrameSyncMorph methodsFor: 'piano rolls' stamp: 'RAA 12/11/2000 22:58'!
justDroppedIntoPianoRoll: pianoRoll event: evt

	"since these morphs handle their own dropping, ignore"! !

!MovieFrameSyncMorph methodsFor: 'access' stamp: 'di 10/11/2000 21:39'!
moviePlayerMorph
 
	^ moviePlayerMorph! !

!MovieFrameSyncMorph methodsFor: 'events' stamp: 'di 10/24/2000 18:22'!
nextSyncEventAfter: index inTrack: track
	| evt |
	index to: track size do:
		[:i | evt _ track at: i.
		((evt morph isMemberOf: self class)
			and: [evt morph moviePlayerMorph == moviePlayerMorph])
			ifTrue: [^ evt]].
	^ nil! !

!MovieFrameSyncMorph methodsFor: 'piano rolls' stamp: 'di 10/25/2000 22:03'!
pauseFrom: scorePlayer

	self relatedPlayer pauseFrom: scorePlayer! !

!MovieFrameSyncMorph methodsFor: 'access' stamp: 'di 10/25/2000 22:02'!
relatedPlayer
 
	^ moviePlayerMorph! !

!MovieFrameSyncMorph methodsFor: 'piano rolls' stamp: 'di 10/25/2000 22:02'!
resetFrom: scorePlayer

	self relatedPlayer resetFrom: scorePlayer! !

!MovieFrameSyncMorph methodsFor: 'piano rolls' stamp: 'di 10/25/2000 22:03'!
resumeFrom: scorePlayer

	self relatedPlayer resumeFrom: scorePlayer! !

Object subclass: #JPEGMovieFile
	instanceVariableNames: 'file movieExtent frameRate frameOffsets currentFrameIndex soundtrackOffsets'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Kernel'!

!JPEGMovieFile commentStamp: '<historical>' prior: 0!
A JPEG movie file consists of a header followed by a sequence of JPEG-compressed images followed by one or more sound tracks. It responds to a subset of the protocol for MPEGFile so that a JPEG movie can be played by MPEGDisplayMorph as if it were an MPEG movie.
!

!JPEGMovieFile class methodsFor: 'movie soundtracks' stamp: 'jm 12/13/2001 21:03'!
addSoundtrack: soundFileName toJPEGMovieNamed: jpegFileName compressionType: compressionTypeString
	"Append the given audio file as a soundtrack the given JPEG movie using the given compression type ('none', 'adpcm3', 'adpcm4', 'adpcm5', 'mulaw', or 'gsm')."
	"Note: While the Squeak JPEG movie format supports multiple soundtracks, the player currently plays only the first soundtrack."

	| snd jpegFile outFile frameCount newFrameOffsets buf inFile newSoundtrackOffsets oldMovieName |
	snd _ StreamingMonoSound onFileNamed: soundFileName.
	jpegFile _ JPEGMovieFile new openFileNamed: jpegFileName.
	outFile _ (FileStream newFileNamed: 'movie.tmp') binary.
	frameCount _ jpegFile videoFrames: 0.

	"write new header"
	self
		writeHeaderExtent: ((jpegFile videoFrameWidth: 0)@(jpegFile videoFrameHeight: 0))
		frameRate: (jpegFile videoFrameRate: 0)
		frameCount: frameCount
		soundtrackCount: (jpegFile soundtrackOffsets size + 1)
		on: outFile.

	"copy frames to new file"
	newFrameOffsets _ Array new: frameCount + 1.
	1 to: frameCount do: [:i |
		newFrameOffsets at: i put: outFile position.
		buf _ jpegFile bytesForFrame: i.
		outFile nextPutAll: buf].
	newFrameOffsets at: frameCount + 1 put: outFile position.

	"copy existing soundtracks, if any, to new file"
	jpegFile soundtrackOffsets size > 0 ifTrue: [
		inFile _ jpegFile fileHandle.
		inFile position: jpegFile soundtrackOffsets first.
		buf _ ByteArray new: 10000.
		[inFile atEnd] whileFalse: [
			buf _ inFile next: buf size into: buf startingAt: 1.
			outFile nextPutAll: buf]].

	"adjust soundtrack offsets for header size increase and add new one:"
	newSoundtrackOffsets _ jpegFile soundtrackOffsets collect: [:n | n + 4].
	newSoundtrackOffsets _ newSoundtrackOffsets copyWith: outFile position.
	snd storeSunAudioOn: outFile compressionType: compressionTypeString.

	"update header:"
	self updateFrameOffsets: newFrameOffsets on: outFile.
	self updateSoundtrackOffsets: newSoundtrackOffsets frameOffsets: newFrameOffsets on: outFile.

	"close files"
	snd closeFile.
	jpegFile closeFile.
	outFile close.

	"replace the old movie with the new version"
	oldMovieName _ (jpegFile fileName copyFrom: 1 to: (jpegFile fileName size - 4)), '.old'.
	FileDirectory default deleteFileNamed: oldMovieName.
	FileDirectory default rename: jpegFile fileName toBe: oldMovieName.
	FileDirectory default rename: 'movie.tmp' toBe: jpegFile fileName.
! !

!JPEGMovieFile class methodsFor: 'movie conversion' stamp: 'jm 1/25/2002 20:46'!
convertFromFolderOfFramesNamed: folderName toJPEGMovieNamed: jpegFileName frameRate: frameRate quality: quality
	"Convert a folder of frames into a JPEG movie. The named folder is assumed to contain only image files, all of the same size, and whose alphabetical order (case-insensitive) is the sequence in which they will appear in in the movie. A useful convention is to make the image files end in zero-padded frame numbers, for example 'frame0001.bmp', 'frame0002.bmp', etc. The image files can be any format readable by Form>fromFileNamed:. The movie frame extent is taken from the first frame file."

	| jpegFile dir fileNames frameCount frameForm frameOffsets |
	(FileDirectory default directoryExists: folderName)
		ifFalse: [^ self inform: 'Folder not found: ', folderName].
	jpegFile _ (FileStream newFileNamed: jpegFileName) binary.
	dir _ FileDirectory default on: folderName.
	fileNames _ self sortedByFrameNumber: dir fileNames.
	frameCount _ fileNames size.
	frameForm _ Form fromFileNamed: (dir fullNameFor: fileNames first).

	"write header"
	self writeHeaderExtent: frameForm extent
		frameRate: frameRate
		frameCount: frameCount
		soundtrackCount: 0
		on: jpegFile.

	"convert and write frames"
	frameOffsets _ Array new: frameCount + 1.
	1 to: frameCount do: [:i |
		frameOffsets at: i put: jpegFile position.
		frameForm _ Form fromFileNamed: (dir fullNameFor: (fileNames at: i)).
		self writeFrame: frameForm on: jpegFile quality: quality displayFlag: true].
	frameOffsets at: (frameCount + 1) put: jpegFile position.
	self updateFrameOffsets: frameOffsets on: jpegFile.

	jpegFile close.
	Display restore.
! !

!JPEGMovieFile class methodsFor: 'movie conversion' stamp: 'jm 12/13/2001 09:38'!
convertMPEGFileNamed: mpegFileName toJPEGMovieNamed: jpegFileName quality: quality
	"Convert the MPEG movie with the given file name into a JPEG movie with the given file name."

	| mpegFile jpegFile soundtrackCount movieExtent frameOffsets soundTrackOffsets |
	(FileDirectory default fileExists: mpegFileName)
		ifFalse: [^ self inform: 'File not found: ', mpegFileName].
	(MPEGFile isFileValidMPEG: mpegFileName)
		ifFalse: [^ self inform: 'Not an MPEG file: ', mpegFileName].
	mpegFile _ MPEGFile openFile: mpegFileName.
	mpegFile fileHandle ifNil: [^ self inform: 'Could not open ', mpegFileName].
	jpegFile _ (FileStream newFileNamed: jpegFileName) binary.

	"write header"
	soundtrackCount _ mpegFile hasAudio ifTrue: [1] ifFalse: [0].
	mpegFile hasVideo
		ifTrue: [
			movieExtent _ (mpegFile videoFrameWidth: 0)@(mpegFile videoFrameHeight: 0).
			self writeHeaderExtent: movieExtent
				frameRate: (mpegFile videoFrameRate: 0)
				frameCount: (mpegFile videoFrames: 0)
				soundtrackCount: soundtrackCount
				on: jpegFile]
		ifFalse: [
			self writeHeaderExtent: 0@0
				frameRate: 0
				frameCount: 0
				soundtrackCount: soundtrackCount
				on: jpegFile].

	"convert and write frames"
	frameOffsets _ self writeFramesFrom: mpegFile on: jpegFile quality: quality.
	self updateFrameOffsets: frameOffsets on: jpegFile.

	"convert and write sound tracks"
	jpegFile position: frameOffsets last.  "store sound tracks after the last frame"
	soundTrackOffsets _ self writeSoundTracksFrom: mpegFile on: jpegFile.
	self updateSoundtrackOffsets: soundTrackOffsets frameOffsets: frameOffsets on: jpegFile.

	mpegFile closeFile.
	jpegFile close.
	Display restore.
! !

!JPEGMovieFile class methodsFor: 'movie conversion' stamp: 'jm 12/13/2001 09:39'!
convertSqueakMovieNamed: squeakMovieFileName toJPEGMovieNamed: jpegFileName quality: quality
	"Convert the Squeak movie with the given file name into a JPEG movie with the given file name."

	| sqMovieFile jpegFile w h d frameCount mSecsPerFrame frameForm bytesPerFrame frameOffsets |
	(FileDirectory default fileExists: squeakMovieFileName)
		ifFalse: [^ self inform: 'File not found: ', squeakMovieFileName].
	sqMovieFile _ (FileStream readOnlyFileNamed: squeakMovieFileName) binary.
	sqMovieFile ifNil: [^ self inform: 'Could not open ', squeakMovieFileName].
	jpegFile _ (FileStream newFileNamed: jpegFileName) binary.

	sqMovieFile nextInt32.  "skip first word"
	w _ sqMovieFile nextInt32.
	h _ sqMovieFile nextInt32.
	d _ sqMovieFile nextInt32.
	frameCount _ sqMovieFile nextInt32.
	mSecsPerFrame _ (sqMovieFile nextInt32) / 1000.0.

	"write header"
	self writeHeaderExtent: w@h
		frameRate: (1000.0 / mSecsPerFrame)
		frameCount: frameCount
		soundtrackCount: 0
		on: jpegFile.

	"convert and write frames"
	frameForm _ Form extent: w@h depth: d.
	bytesPerFrame _ 4 + (frameForm bits size * 4).
	frameOffsets _ Array new: frameCount + 1.
	1 to: frameCount do: [:i |
		frameOffsets at: i put: jpegFile position.
		sqMovieFile position: 128 + ((i - 1) * bytesPerFrame) + 4.
		sqMovieFile next: frameForm bits size into: frameForm bits startingAt: 1.
		frameForm display.
		self writeFrame: frameForm on: jpegFile quality: quality displayFlag: false].
	frameOffsets at: (frameCount + 1) put: jpegFile position.
	self updateFrameOffsets: frameOffsets on: jpegFile.

	sqMovieFile close.
	jpegFile close.
	Display restore.
! !

!JPEGMovieFile class methodsFor: 'movie creation-private' stamp: 'jm 1/25/2002 21:08'!
extractFrameNumberFrom: aString
	"Answer the integer frame number from the given file name string. The frame number is assumed to be the last contiguous sequence of digits in the given string. For example, 'take2 005.jpg' is frame 5 of the sequence 'take2'."
	"Assume: The given string contains at least one digit."

	| end start |
	end _ aString size.
	[(aString at: end) isDigit not] whileTrue: [end _ end - 1].
	start _ end.
	[(start > 1) and: [(aString at: start - 1) isDigit]] whileTrue: [start _ start - 1].
	^ (aString copyFrom: start to: end) asNumber
! !

!JPEGMovieFile class methodsFor: 'testing' stamp: 'jm 11/15/2001 14:56'!
isJPEGMovieFile: fileName
	"Answer true if the file with the given name appears to be a JPEG movie file."

	| f tag |
	(FileDirectory default fileExists: fileName) ifFalse: [^ false].
	f _ (FileStream readOnlyFileNamed: fileName) binary.
	tag _ (f next: 10) asString.
	f close.
	^ tag = 'JPEG Movie'
! !

!JPEGMovieFile class methodsFor: 'movie soundtracks' stamp: 'jm 12/13/2001 21:03'!
removeSoundtrackFromJPEGMovieNamed: jpegFileName
	"Remove all soundtracks from the JPEG movie with the given name."

	| jpegFile outFile frameCount newFrameOffsets buf oldMovieName |
	jpegFile _ JPEGMovieFile new openFileNamed: jpegFileName.
	outFile _ (FileStream newFileNamed: 'movie.tmp') binary.
	frameCount _ jpegFile videoFrames: 0.

	"write new header"
	self
		writeHeaderExtent: ((jpegFile videoFrameWidth: 0)@(jpegFile videoFrameHeight: 0))
		frameRate: (jpegFile videoFrameRate: 0)
		frameCount: frameCount
		soundtrackCount: 0
		on: outFile.

	"copy frames to new file"
	newFrameOffsets _ Array new: frameCount + 1.
	1 to: frameCount do: [:i |
		newFrameOffsets at: i put: outFile position.
		buf _ jpegFile bytesForFrame: i.
		outFile nextPutAll: buf].
	newFrameOffsets at: frameCount + 1 put: outFile position.

	"update header:"
	self updateFrameOffsets: newFrameOffsets on: outFile.

	"close files"
	jpegFile closeFile.
	outFile close.

	"replace the old movie with the new version"
	oldMovieName _ (jpegFile fileName copyFrom: 1 to: (jpegFile fileName size - 4)), '.old'.
	FileDirectory default deleteFileNamed: oldMovieName.
	FileDirectory default rename: jpegFile fileName toBe: oldMovieName.
	FileDirectory default rename: 'movie.tmp' toBe: jpegFile fileName.
! !

!JPEGMovieFile class methodsFor: 'movie creation-private' stamp: 'jm 2/3/2002 10:14'!
sortedByFrameNumber: fileNames
	"Sort the given collection of fileNames by frame number. The frame number is the integer value of the last contiguous sequence of digits in the file name. Omit filenames that do not contain at least one digit; this helps filter out extraneous non-frame files such as the invisible 'Icon' file that may be inserted by some file servers."

	| filtered pairs |
	"select the file names contain at least one digit"
	filtered _ fileNames select: [:fn | fn anySatisfy: [:c | c isDigit]].

	"make array of number, name pairs"
	pairs _ filtered asArray collect: [:fn |
		Array with: (self extractFrameNumberFrom: fn) with: fn].

	"sort the pairs, then answer a collection containing the second element of every pair"
	pairs sort: [:p1 :p2 | p1 first < p2 first].
	^ pairs collect: [:p | p last].
! !

!JPEGMovieFile class methodsFor: 'movie creation-private' stamp: 'jm 11/17/2001 08:05'!
updateFrameOffsets: frameOffsets on: aBinaryStream
	"Update the JPEG movie file header on the given stream with the given collection of frame offsets."

	aBinaryStream position: 22.
	frameOffsets do: [:offset | aBinaryStream uint32: offset].
! !

!JPEGMovieFile class methodsFor: 'movie creation-private' stamp: 'jm 11/17/2001 07:40'!
updateSoundtrackOffsets: soundtrackOffsetList frameOffsets: frameOffsets on: aBinaryStream
	"Update the JPEG movie file header on the given stream with the given sequence of sound track offsets."

	aBinaryStream position: 22 + (4 * frameOffsets size).
	aBinaryStream uint16: soundtrackOffsetList size.
	soundtrackOffsetList do: [:offset | aBinaryStream uint32: offset].
! !

!JPEGMovieFile class methodsFor: 'movie creation-private' stamp: 'jm 11/25/2001 14:20'!
writeFrame: aForm on: aBinaryStream quality: quality displayFlag: displayFlag
	"Compress and the given Form on the given stream and answer its offset. If displayFlag is true, show the result of JPEG compression on the display."

	| offset compressed outForm |
	offset _ aBinaryStream position.
	compressed _ JPEGReadWriter2 new compress: aForm quality: quality.
	displayFlag ifTrue: [  "show decompressed frame"
		outForm _ (JPEGReadWriter2 on: (ReadStream on: compressed)) nextImage.
		outForm display].
	aBinaryStream nextPutAll: compressed.
	^ offset
! !

!JPEGMovieFile class methodsFor: 'movie creation-private' stamp: 'jm 11/25/2001 14:23'!
writeFramesFrom: mpegFile on: aBinaryStream quality: quality
	"Write the frames of the given MPEG movie on the given stream at the given JPEG quality level. Answer a collection of frame offsets. The size of this collection is one larger than the number of frames; it's final entry is the stream position just after the final frame. The byte count for any frame can thus be computed as the difference between two adjacent offsets."

	| frameCount frameOffsets frameForm |
	mpegFile hasVideo ifFalse: [^ Array with: aBinaryStream position].
	frameCount _ mpegFile videoFrames: 0.
	frameOffsets _ OrderedCollection new: frameCount + 1.
	frameForm _ Form
		extent: (mpegFile videoFrameWidth: 0)@(mpegFile videoFrameHeight: 0)
		depth: 32.

	[(mpegFile videoGetFrame: 0) < (mpegFile videoFrames: 0)] whileTrue: [
		frameOffsets addLast: aBinaryStream position.
		mpegFile videoReadFrameInto: frameForm stream: 0.
		self writeFrame: frameForm on: aBinaryStream quality: quality displayFlag: true].

	frameOffsets addLast: aBinaryStream position.  "add final offset"
	^ frameOffsets
! !

!JPEGMovieFile class methodsFor: 'movie creation-private' stamp: 'jm 11/17/2001 08:01'!
writeHeaderExtent: movieExtent frameRate: frameRate frameCount: frameCount soundtrackCount: soundtrackCount on: aBinaryStream
	"Write a header on the given stream for a JPEG movie file with the given specifications. Leave the stream positioned at the start of the first movie frame."

	| offsetCount |
	aBinaryStream position: 0.
	aBinaryStream nextPutAll: ('JPEG Movie') asByteArray.
	aBinaryStream uint16: movieExtent x.
	aBinaryStream uint16: movieExtent y.
	aBinaryStream uint32: (frameRate * 10000) rounded.
	offsetCount _ frameCount + 1.
	aBinaryStream uint32: offsetCount.
	aBinaryStream skip: (offsetCount * 4).  "leave room for frame offsets"
	aBinaryStream uint16: soundtrackCount.
	aBinaryStream skip: (soundtrackCount * 4).  "leave room for sound track offsets"
! !

!JPEGMovieFile class methodsFor: 'movie creation-private' stamp: 'jm 11/25/2001 16:55'!
writeSoundTracksFrom: mpegFile on: aBinaryStream
	"Convert and write the sound tracks from the given MPEG file to given stream. Answer a collection of sound track offsets."
	"Details: Currently converts at most one sound track; only the left channel of a stereo movie will be converted."

	| soundtrackCount soundTrackOffsets snd |
	soundtrackCount _ mpegFile hasAudio ifTrue: [1] ifFalse: [0].
	soundTrackOffsets _ Array new: soundtrackCount.
	1 to: soundtrackCount do: [:i |
		soundTrackOffsets at: i put: aBinaryStream position.
		snd _ mpegFile audioPlayerForChannel: i.
		snd storeSunAudioOn: aBinaryStream compressionType: 'mulaw'.
		snd closeFile].
	^ soundTrackOffsets
! !

!JPEGMovieFile methodsFor: 'audio' stamp: 'jm 11/17/2001 09:49'!
audioPlayerForChannel: anInteger
	"Answer a streaming sound for playing the audio channel with the given index."

	((anInteger >= 1) & (anInteger <= soundtrackOffsets size)) ifFalse: [^ nil].
	^ StreamingMonoSound
		onFileNamed: file fullName
		headerStart: (soundtrackOffsets at: anInteger)
! !

!JPEGMovieFile methodsFor: 'private' stamp: 'jm 11/15/2001 08:02'!
bytesForFrame: frameIndex
	"Answer a ByteArray containing the encoded bytes for the frame with the given index. Answer nil if the index is out of range or if my file is not open."

	frameIndex < 1 ifTrue: [^ nil].
	frameIndex >= frameOffsets size ifTrue: [^ nil].
	file ifNil: [^ nil].
	file closed ifTrue: [file ensureOpen; binary].
	file position: (frameOffsets at: frameIndex).
	^ file next: (frameOffsets at: frameIndex + 1) - (frameOffsets at: frameIndex)
! !

!JPEGMovieFile methodsFor: 'file ops' stamp: 'jm 11/14/2001 14:08'!
closeFile
	"Close my file stream."

	file ifNotNil: [file close].
! !

!JPEGMovieFile methodsFor: 'file ops' stamp: 'jm 11/14/2001 14:13'!
fileHandle
	"Answer my file, or nil if the file is not open."

	file ifNil: [^ nil].
	file closed ifTrue: [^ nil].
	^ file
! !

!JPEGMovieFile methodsFor: 'file ops' stamp: 'jm 11/15/2001 07:59'!
fileName
	"Answer the name of my file."

	file ifNil: [^ ''].
	^ file fullName
! !

!JPEGMovieFile methodsFor: 'audio' stamp: 'jm 11/16/2001 17:03'!
hasAudio
	"Answer true if this movie has at least one sound track."

	^ soundtrackOffsets size > 0
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/20/2001 16:08'!
hasVideo
	"Answer true if I have one or more frames."

	^ frameOffsets size > 1  "note: the empty movie still has one frameOffset"
! !

!JPEGMovieFile methodsFor: 'initialization' stamp: 'jm 11/15/2001 08:13'!
initialize

	file _ nil.
	frameOffsets _ #().
	currentFrameIndex _ 1.
! !

!JPEGMovieFile methodsFor: 'file ops' stamp: 'jm 2/11/2002 13:09'!
openFile: fileName
	"For compatability with MPEGFile."

	self openFileNamed: fileName.
! !

!JPEGMovieFile methodsFor: 'file ops' stamp: 'jm 2/11/2002 13:30'!
openFileNamed: fileName
	"Open the JPEG movie file with the given name."

	file ifNotNil: [file finalize].
	file _ nil.
	(FileDirectory default fileExists: fileName) ifFalse: [^ self].
	file _ (FileStream readOnlyFileNamed: fileName) binary.
	self readHeader.
	currentFrameIndex _ 1.
! !

!JPEGMovieFile methodsFor: 'private' stamp: 'jm 11/16/2001 17:01'!
readHeader
	"Read a JPEG movie header file."
	"Details: The file structures is:
		<header, including sequence frame offsets>
		<sequence of JPEG compressed images>
		<optional soundtracks>"

	| tag w h frameOffsetCount soundtrackCount |
	file position: 0.
	tag _ (file next: 10) asString.
	tag = 'JPEG Movie' ifFalse: [self error: 'not a JPEG movie file'].
	w _ file uint16.
	h _ file uint16.
	movieExtent _ w @ h.
	frameRate _ file uint32 / 10000.0.
	frameOffsetCount _ file uint32.
	frameOffsets _ Array new: frameOffsetCount.
	1 to: frameOffsetCount do: [:i | frameOffsets at: i put: file uint32].
	soundtrackCount _ file uint16.
	soundtrackOffsets _ Array new: soundtrackCount.
	1 to: soundtrackCount do: [:i | soundtrackOffsets at: i put: file uint32].
! !

!JPEGMovieFile methodsFor: 'private' stamp: 'jm 12/13/2001 19:14'!
soundtrackOffsets
	"Answer the offsets for my soundtracks."

	^ soundtrackOffsets
! !

!JPEGMovieFile methodsFor: 'private' stamp: 'jm 11/27/2001 10:23'!
testPlay
	"Performance benchmark. Decompress and display all my frames. Answer the frame rate achieved in frames/second. No sound is played."

	| frameForm frameCount t |
	frameForm _ Form extent: movieExtent depth: (Display depth max: 16).
	frameCount _ self videoFrames: 0.
	self videoSetFrame: 1 stream: 0.
	t _ [
		frameCount timesRepeat: [
			self videoReadFrameInto: frameForm stream: 0.
			frameForm display].
	] timeToRun.
	^ ((1000.0 * frameCount) / t) roundTo: 0.01
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/14/2001 14:18'!
videoDropFrames: skipCount stream: streamIndex
	"Advance the index of the current frame by the given number of frames."

	self videoSetFrame: currentFrameIndex + skipCount stream: streamIndex.
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/15/2001 08:08'!
videoFrameHeight: ignored
	"Answer the height of this movie."

	^ movieExtent y
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/15/2001 07:37'!
videoFrameRate: ignored
	"Answer the frame rate of this movie."

	^ frameRate
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/15/2001 08:08'!
videoFrameWidth: ignored
	"Answer the width of this movie."

	^ movieExtent x
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/15/2001 07:34'!
videoFrames: ignored
	"Answer the number of video frames in this movie."

	^ (frameOffsets size - 1) max: 0
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/14/2001 14:05'!
videoGetFrame: ignored
	"Answer the index of the current frame, or zero if the movie has no frames."

	^ currentFrameIndex
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/20/2001 11:02'!
videoReadFrameInto: aForm stream: aStream
	"Read the next frame into the given 16-bit or 32-bit Form."

	| compressedBytes |
	compressedBytes _ self bytesForFrame: currentFrameIndex.
	compressedBytes ifNil: [^ self].
	JPEGReadWriter2 new uncompress: compressedBytes into: aForm.
	currentFrameIndex _ (currentFrameIndex + 1) min: (frameOffsets size - 1).
! !

!JPEGMovieFile methodsFor: 'video' stamp: 'jm 11/15/2001 08:10'!
videoSetFrame: newIndex stream: ignored
	"Set the index of the current frame."

	currentFrameIndex _ (newIndex asInteger max: 1) min: (frameOffsets size - 1).
! !

Object subclass: #MPEGFile
	instanceVariableNames: 'pathToFile fileBits fileIndex endianness'
	classVariableNames: 'Registry'
	poolDictionaries: ''
	category: 'Movies-Kernel'!

!MPEGFile commentStamp: '<historical>' prior: 0!
* An interface to LibMPEG3
 * Author: Adam Williams <broadcast@earthling.net>
 * Page: heroine.linuxbox.com
 *
 * Changed for Squeak to work with Squeak and to work on the Macintosh
 * Sept 2000, by John M McIntosh johnmci@smalltalkconsulting.com
 * The smalltalk code and the C code it produces is released under the 
 * Squeak licence. The libmpeg3 C code is co-licenced under either the Squeak licence or
 * the GNU LGPL
!

!MPEGFile class methodsFor: 'testing' stamp: 'JMM 9/18/2000 14:28'!
isFileValidMPEG: path
	^self primFileValidMPEG: path! !

!MPEGFile class methodsFor: 'instance creation' stamp: 'nk 7/30/2004 21:50'!
openFile: aPath 
	^self new openFile: aPath! !

!MPEGFile class methodsFor: 'primitives' stamp: 'JMM 9/18/2000 14:27'!
primFileValidMPEG: aPath
	"Check to see if the file is valid"
	<primitive: 'primitiveMPEG3CheckSig' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile class methodsFor: 'registry' stamp: 'JMM 9/17/2000 23:56'!
register: anObject
	WeakArray isFinalizationSupported ifFalse:[^anObject].
	self registry add: anObject! !

!MPEGFile class methodsFor: 'registry' stamp: 'JMM 9/17/2000 23:56'!
registry
	WeakArray isFinalizationSupported ifFalse:[^nil].
	^Registry isNil
		ifTrue:[Registry := WeakRegistry new]
		ifFalse:[Registry].! !

!MPEGFile class methodsFor: 'registry' stamp: 'JMM 9/17/2000 23:56'!
unregister: anObject
	WeakArray isFinalizationSupported ifFalse:[^anObject].
	self registry remove: anObject ifAbsent:[]! !

!MPEGFile methodsFor: 'audio' stamp: 'jm 11/17/2001 08:18'!
audioChannels: aStream
	"Returns -1 if error, otherwise returns audioChannels for stream aStream"
	self hasAudio ifFalse: [^ 0].
	^[self primAudioChannels: self fileHandle stream: aStream] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 11/21/2000 13:13'!
audioGetSample: aStream
	"Returns number of current sample, or -1 if error"
	self hasAudio ifFalse: [^-1].
	^[(self primGetSample: self fileHandle stream: aStream) asInteger] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'audio' stamp: 'jm 11/17/2001 09:36'!
audioPlayerForChannel: channelNumber
	"Answer a streaming sound for playing the audio channel with the given index."
	"Note: The MP3 player can not yet isolate a single channel from a multi-channel audio stream."

	^ StreamingMP3Sound new initMPEGFile: self streamIndex: 0
! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 11/21/2000 13:13'!
audioReReadBuffer: aBuffer stream: aStreamNumber channel: aChannelNumber
	"Used to read other channels after first ReadBuffer 
	Returns -1 if error, otherwise 0"
	self hasAudio ifFalse: [^-1].
	^[self audioReReadBuffer: aBuffer stream: aStreamNumber channel: aChannelNumber samples: (aBuffer size * aBuffer bytesPerElement // 2)] on: Error do: [-1]! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 11/21/2000 13:13'!
audioReReadBuffer: aBuffer stream: aStreamNumber channel: aChannelNumber samples: aSampleNumber
	"Used to read other channels after first ReadBuffer 
	Returns -1 if error, otherwise 0
	Note this call requires passing in the samples to read, ensure you get the number right"
	self hasAudio ifFalse: [^-1].
	^[self primAudioReReadBuffer: self fileHandle  buffer: aBuffer channel: aChannelNumber samples: aSampleNumber stream: aStreamNumber] on: Error do: [-1]! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 11/21/2000 13:13'!
audioReadBuffer: aBuffer stream: aStreamNumber channel: aChannelNumber 
	"Returns -1 if error, otherwise 0"
	self hasAudio ifFalse: [^-1].
	^[self audioReadBuffer: aBuffer stream: aStreamNumber channel: aChannelNumber samples: (aBuffer size* aBuffer bytesPerElement)//2] on: Error do: [-1]! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 11/21/2000 13:13'!
audioReadBuffer: aBuffer stream: aStreamNumber channel: aChannelNumber samples: aSampleNumber
	"Returns -1 if error, otherwise 0
	Note this call requires passing in the samples to read, ensure you get the number right"
	self hasAudio ifFalse: [^-1].
	^[self primAudioReadBuffer: self fileHandle  buffer: aBuffer channel: aChannelNumber samples: aSampleNumber stream: aStreamNumber] on: Error do: [-1]! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 11/21/2000 13:13'!
audioSampleRate: aStream
	"Returns sample rate, or -1 if error"
	self hasAudio ifFalse: [^-1].
	^[self primSampleRate: self fileHandle stream: aStream] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 11/21/2000 13:13'!
audioSamples: aStream
	"Returns -1 if error, 
	otherwise returns audioSamples for stream aStream"
	self hasAudio ifFalse: [^-1].
	^[(self primAudioSamples: self fileHandle stream: aStream) asInteger] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 11/21/2000 13:13'!
audioSetSample: aNumber stream: aStream
	"Set number of targeted sample, returns 0 if ok, -1 if failure"
	self hasAudio ifFalse: [^-1].
	^[self primSetSample: self fileHandle sample: aNumber asFloat stream: aStream] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'initialize-release' stamp: 'JMM 9/20/2000 01:59'!
closeFile
	self finalize.! !

!MPEGFile methodsFor: 'converting' stamp: 'dgd 2/16/2004 14:19'!
convertToSqueakMovieFileNamed: fileName 
	"convert the receiver to a squeak-format movie"
	" 
	(MPEGFile openFile:
	'/H/squeak/Small-Land/Demo/media/mazinger_z_spanish_op.mpg') 
	convertToSqueakMovieFileNamed: 'MazingerZ.squeakmovie' 
	"
	| movieFile max w h d frameBuffer |
	movieFile := FileStream newFileNamed: fileName.
	[movieFile binary.
	"no idea what goes here..."
	movieFile nextInt32Put: 0.
	movieFile nextInt32Put: (w := self videoFrameWidth: 0).
	movieFile nextInt32Put: (h := self videoFrameHeight: 0).
	"Depth of form data stored"
	"we really don't know but try to preserve some space"
	movieFile nextInt32Put: (d := 16).
	movieFile nextInt32Put: (max := self videoFrames: 0).
	"min: 100"
	movieFile nextInt32Put: (1000 * 1000
			/ (self videoFrameRate: 0)) rounded.
	"Padding?"
	movieFile
		nextPutAll: (ByteArray new: 128 - movieFile position).
	frameBuffer := Form extent: w @ h depth: d.
	self videoSetFrame: 1 stream: 0.
	'Converting movie...'
		displayProgressAt: Sensor cursorPoint
		from: 1
		to: max
		during: [:bar | 1
				to: max
				do: [:i | 
					bar value: i.
					self videoReadFrameInto: frameBuffer stream: 0.
					frameBuffer display.
					movieFile nextInt32Put: i.
					movieFile nextPutAll: frameBuffer bits]]]
		ensure: [movieFile close]! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 9/20/2000 01:57'!
endOfAudio: aStream
	"Returns true if end of Audio"
	self hasAudio ifFalse: [^true].
	^self primEndOfAudio: self fileHandle stream: aStream
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 9/20/2000 01:54'!
endOfVideo: aStream
	"Returns true if end of video"
	self hasVideo ifFalse: [^true].
	^self primEndOfVideo: self fileHandle stream: aStream
! !

!MPEGFile methodsFor: 'access' stamp: 'yo 7/2/2004 15:58'!
endianness
	^endianness isNil 
		ifTrue: [endianness _ SmalltalkImage current endianness] 
		ifFalse: [endianness]! !

!MPEGFile methodsFor: 'access' stamp: 'JMM 9/20/2000 19:04'!
fileHandle
	(Smalltalk externalObjects at: fileIndex ifAbsent: [^nil]) == fileBits 
		ifTrue: [^fileBits]
		ifFalse: [^nil].
! !

!MPEGFile methodsFor: 'access' stamp: 'JMM 9/18/2000 18:38'!
fileName
	^pathToFile! !

!MPEGFile methodsFor: 'file ops' stamp: 'JMM 9/20/2000 02:05'!
finalize
	self fileHandle notNil ifTrue: [self primFileClose: self fileHandle].
	self fileHandle = fileBits ifTrue: [Smalltalk unregisterExternalObject: fileIndex].
	fileBits _ nil.	
	fileIndex _ 0.! !

!MPEGFile methodsFor: 'access' stamp: 'JMM 9/20/2000 01:54'!
getPercentage
	"Return current location by percentage, 0.0-1.0"
	^self primGetPercentage: self fileHandle ! !

!MPEGFile methodsFor: 'access' stamp: 'JMM 9/29/2000 19:28'!
getTOC: timecode doStreams: streams
	| buffer |
	
	buffer _ String new: 64*1024+1.
	self primGenerateToc: self fileHandle useSearch: timecode doStreams: streams buffer: buffer.
	^buffer! !

!MPEGFile methodsFor: 'access' stamp: 'JMM 9/20/2000 01:56'!
getTimeCode
	"Return time code, (float) "
	^self primGetTime: self fileHandle! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 9/20/2000 01:56'!
hasAudio
	"Returns true if file has audio"
	^self primHasAudio: self fileHandle ! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 9/20/2000 01:54'!
hasVideo
	"Returns true if file has video"
	^self primHasVideo: self fileHandle ! !

!MPEGFile methodsFor: 'initialize-release' stamp: 'JMM 9/20/2000 01:57'!
openFile: aPath
	pathToFile _ aPath.
	fileBits _ self primFileOpen: aPath.
	fileBits notNil ifTrue: 
		[fileIndex _ Smalltalk registerExternalObject: fileBits.
		self register.]
	! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 14:19'!
primAudioChannels: aHandle stream: aStream
	<primitive: 'primitiveMPEG3AudioChannels' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/19/2000 13:35'!
primAudioReReadBuffer: aFileHandle  buffer: aBuffer channel: aChannel samples: aSampleNumber stream: aStreamNumber
	<primitive: 'primitiveMPEG3ReReadAudio' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/19/2000 13:31'!
primAudioReadBuffer: aFileHandle  buffer: aBuffer channel: aChannel samples: aSampleNumber stream: aStreamNumber
	<primitive: 'primitiveMPEG3ReadAudio' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 14:23'!
primAudioSamples: aHandle stream: aStream
	<primitive: 'primitiveMPEG3AudioSamples' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 17:54'!
primDropFrame: aHandle frame: aNumberOfFrames stream: aStream
	<primitive: 'primitiveMPEG3DropFrames' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 14:50'!
primEndOfAudio: aHandle stream: aStream
	<primitive: 'primitiveMPEG3EndOfAudio' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 14:51'!
primEndOfVideo: aHandle stream: aStream
	<primitive: 'primitiveMPEG3EndOfVideo' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 13:58'!
primFileClose: aHandle
	"Close the file"
	<primitive: 'primitiveMPEG3Close' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 03:56'!
primFileOpen: aPath
	"Open the file"
	<primitive: 'primitiveMPEG3Open' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 14:52'!
primFrameRate: aHandle stream: aStream
	<primitive: 'primitiveMPEG3FrameRate' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/29/2000 17:29'!
primGenerateToc: fileHandle useSearch: timecode doStreams: streams buffer: aString
	<primitive: 'primitiveMPEG3GenerateToc' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:11'!
primGetFrame: aHandle stream: aStream
	<primitive: 'primitiveMPEG3GetFrame' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:37'!
primGetPercentage: aHandle
	<primitive: 'primitiveMPEG3TellPercentage' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:12'!
primGetSample: aHandle stream: aStream
	<primitive: 'primitiveMPEG3GetSample' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:15'!
primGetTime: aFileHandle
	<primitive: 'primitiveMPEG3GetTime' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 14:59'!
primHasAudio: aHandle
	<primitive: 'primitiveMPEG3HasAudio' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 14:59'!
primHasVideo: aHandle
	<primitive: 'primitiveMPEG3HasVideo' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:17'!
primPreviousFrame: aHandle stream: aStream
	<primitive: 'primitiveMPEG3PreviousFrame' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:19'!
primSampleRate: aHandle stream: aStream
	<primitive: 'primitiveMPEG3SampleRate' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:23'!
primSeekPercentage: aHandle percentage: aNumber
	<primitive: 'primitiveMPEG3SeekPercentage' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:27'!
primSetCPUs: aHandle number: aNumber
	"Not support on the macintosh below OS X"
	<primitive: 'primitiveMPEG3SetCpus' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:29'!
primSetFrame: aHandle frame: aFrameNumber stream: aStream
	<primitive: 'primitiveMPEG3SetFrame' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:31'!
primSetMMX: aFileHandle useMMX: aValue  
	<primitive: 'primitiveMPEG3SetMmx' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:35'!
primSetSample: aHandle sample: aSampleNumber stream: aStream
	<primitive: 'primitiveMPEG3SetSample' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:39'!
primTotalAudioStreams: aFileHandle 
	<primitive: 'primitiveMPEG3TotalAStreams' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:40'!
primTotalVideoStreams: aFileHandle 
	<primitive: 'primitiveMPEG3TotalVStreams' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:41'!
primVideoFrames: aFileHandle  stream: aStream
	<primitive: 'primitiveMPEG3VideoFrames' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 15:41'!
primVideoHeight: aFileHandle  stream: aStream
	<primitive: 'primitiveMPEG3VideoHeight' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/19/2000 13:28'!
primVideoReadNextFrameFor: aFileHandle into: aFormBuffer x: x y: y width: width height: height outWidth: aTargetWidth outHeight: aTargetHeight colorModel: colorModel stream: aStream bytesPerRow: aByteCount
	<primitive: 'primitiveMPEG3ReadFrame' module: 'Mpeg3Plugin'>

! !

!MPEGFile methodsFor: 'primitives' stamp: 'JMM 9/18/2000 16:35'!
primVideoWidth: aFileHandle  stream: aStream
	<primitive: 'primitiveMPEG3VideoWidth' module: 'Mpeg3Plugin'>
	self primitiveFailed! !

!MPEGFile methodsFor: 'private' stamp: 'JMM 9/17/2000 23:58'!
register
	^self class register: self! !

!MPEGFile methodsFor: 'access' stamp: 'JMM 9/20/2000 01:53'!
seekPercentage: aFloat
	self primSeekPercentage: self fileHandle percentage: aFloat asFloat ! !

!MPEGFile methodsFor: 'access' stamp: 'JMM 9/20/2000 01:56'!
setMMX: aValue
	" true is set, false is off. May not be supported "
	self primSetMMX: self fileHandle useMMX: aValue  ! !

!MPEGFile methodsFor: 'audio' stamp: 'JMM 9/20/2000 01:53'!
totalAudioStreams
	"Returns total number of audio streams"
	^self primTotalAudioStreams: self fileHandle 
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 9/20/2000 01:59'!
totalVideoStreams
	"Returns total number of video streams"
	^self primTotalVideoStreams: self fileHandle 
! !

!MPEGFile methodsFor: 'private' stamp: 'JMM 9/17/2000 23:58'!
unregister
	^self class unregister: self! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 11/21/2000 13:14'!
videoDropFrames: aNumberOfFrames stream: aStream
	"Returns -1 if setFrame failed"
	self hasVideo ifFalse: [^-1].
	^[self primDropFrame: self fileHandle frame: aNumberOfFrames stream: aStream] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 11/21/2000 13:14'!
videoFrameHeight: aStream
	"Returns video frame height, -1 if error "
	self hasVideo ifFalse: [^-1].
	^[self primVideoHeight: self fileHandle stream: aStream] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 11/21/2000 13:14'!
videoFrameRate: aStream
	"Returns video frame rate (float), -1 if error"
	self hasVideo ifFalse: [^-1].
	^[self primFrameRate: self fileHandle stream: aStream] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 11/21/2000 13:14'!
videoFrameWidth: aStream
	"Returns video frame width, -1 if error"
	self hasVideo ifFalse: [^-1].
	^[self primVideoWidth: self fileHandle stream: aStream] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 9/20/2000 01:58'!
videoFrames: aStream
	"Total number of frames" 
	^(self primVideoFrames: self fileHandle stream: aStream) asInteger
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 11/21/2000 13:14'!
videoGetFrame: aStream
	"Returns frame number, or -1 if error"
	self hasVideo ifFalse: [^-1].
	^[(self primGetFrame: self fileHandle stream: aStream) asInteger] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 11/21/2000 13:15'!
videoPreviousFrame: aStream
	"Returns 0 if ok"
	self hasVideo ifFalse: [^-1].
	^[self primPreviousFrame: self fileHandle stream: aStream] on: Error do: [-1]
! !

!MPEGFile methodsFor: 'video' stamp: 'jm 11/16/2001 07:53'!
videoReadFrameInto: aForm stream: aStream
	"Read the next video frame from the given stream into the given 16- or 32-bit Form. The movie frame will be scaled to fit the Form if necessary."

	| colorModel bytesPerRow |
	((aForm depth = 16) | (aForm depth = 32)) ifFalse: [self error: 'must use 16- or 32-bit Form'].
	aForm depth = 16
		ifTrue: [
			colorModel _ self endianness = #big ifTrue: [14] ifFalse: [16].
			bytesPerRow _ 2 * (aForm width roundUpTo: 2)]
		ifFalse: [
			colorModel _ self endianness = #big ifTrue: [13] ifFalse: [1].
			bytesPerRow _ 4 * aForm width].
 	^ self
		videoReadNextFrameInto: aForm bits
		x: 0 y: 0
		width: (self videoFrameWidth: aStream)
		height: (self videoFrameHeight: aStream)
		outWidth: aForm width
		outHeight: aForm height
		colorModel: colorModel
		stream: aStream
		bytesPerRow: bytesPerRow
! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 11/21/2000 13:15'!
videoReadNextFrameInto: aFormBuffer x: x y: y width: width height: height outWidth: aTargetWidth outHeight: aTargetHeight colorModel: colorModel stream: aStream bytesPerRow: aByteCount
	"return nonZero if failure "

	self hasVideo ifFalse: [^-1].
	^[self primVideoReadNextFrameFor: self fileHandle into: aFormBuffer x: x y: y width: width height: height outWidth: aTargetWidth outHeight: aTargetHeight colorModel: colorModel stream: aStream bytesPerRow: aByteCount] on: Error do: [-1]

"/* Supported color models for mpeg3_read_frame */
#define MPEG3_RGB565 2
#define MPEG3_RGB555 14  //JMM for mac
#define MPEG3_RGBI555 16  //SVP for intel
#define MPEG3_BGR888 0
#define MPEG3_BGRA8888 1
#define MPEG3_RGB888 3
#define MPEG3_RGBA8888 4  
#define MPEG3_ARGB8888 13  //JMM for mac
#define MPEG3_RGBA16161616 5

/* Color models for the 601 to RGB conversion */
/* 601 not implemented for scalar code */
#define MPEG3_601_RGB565 11
#define MPEG3_601_RGB555 15 //JMM for Squeak
#define MPEG3_601_RGBI555 17 //SVP for intel
#define MPEG3_601_BGR888 7
#define MPEG3_601_BGRA8888 8
#define MPEG3_601_RGB888 9
#define MPEG3_601_RGBA8888 10
#define MPEG3_601_ARGB8888 12 //JMM for Squeak
"! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 9/20/2000 01:58'!
videoSetCPUs: aNumber
	self primSetCPUs: self fileHandle number: aNumber! !

!MPEGFile methodsFor: 'video' stamp: 'JMM 11/21/2000 13:15'!
videoSetFrame: aFrameNumber stream: aStream
	"Returns -1 if setFrame failed"
	self hasVideo ifFalse: [^-1].
	^[self primSetFrame: self fileHandle frame: aFrameNumber asFloat stream: aStream] on: Error do: [-1]
! !

Object subclass: #MPEGPlayer
	instanceVariableNames: 'external form startTime clockBias frameRate lastDelay noSound sampleRate audioPlayerProcess videoPlayerProcess playerProcessPriority soundQueue timeCheck semaphoreForSound errorForSoundStart morph volume buffer isBuffer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Kernel'!

!MPEGPlayer commentStamp: '<historical>' prior: 0!
V1.01 johnmci@smalltalkconsulting.com Nov 8th 2000
A Simple MPEG Player for playing MPEG3 audio or video 

| foo |
foo _ MpegPlayer playFile: 'my.mpg'.
foo playStream: 0. "To play both audio and video, stream #0 "

foo playAudioStream: 0 "To play audio stream".
foo playVideoStream: 0 "To play video stream"
foo playStreamWaitTilDone: 0 "To play audio/video as currrent process"

or 
	| foo |
	foo _ MPEGPlayer playFile: 'my.mpg' onForm: Display.
	foo playStream: 0.
	To play full screen.
!

!MPEGPlayer class methodsFor: 'instance creation' stamp: 'JMM 1/20/2006 18:16'!
playBuffer: aBuffer onForm: aForm
	^self new initializeWithBuffer: aBuffer morph: aForm! !

!MPEGPlayer class methodsFor: 'instance creation' stamp: 'JMM 1/20/2006 18:06'!
playBuffer: aBuffer onMorph: aMorph
	^self new initializeWithBuffer: aBuffer morph: aMorph! !

!MPEGPlayer class methodsFor: 'instance creation' stamp: 'JMM 9/18/2000 19:02'!
playFile: aPath
	^self new initializeWithFileName: aPath ! !

!MPEGPlayer class methodsFor: 'instance creation' stamp: 'JMM 9/18/2000 18:32'!
playFile: aPath onForm: aForm
	^self new initializeWithFileName: aPath form: aForm! !

!MPEGPlayer class methodsFor: 'instance creation' stamp: 'JMM 10/17/2000 23:19'!
playFile: aPath onMorph: aMorph
	^self new initializeWithFileName: aPath morph: aMorph! !

!MPEGPlayer class methodsFor: 'file suffixes' stamp: 'bkv 11/21/2002 15:28'!
registeredAudioFileSuffixes
    "Answer the file extensions for which MPEGPlayer registers audio services with FileList."
     "MPEGPlayer registeredAudioFileSuffixes"

     ^{ 'mp3'.  }

	! !

!MPEGPlayer class methodsFor: 'file suffixes' stamp: 'bkv 11/21/2002 11:14'!
registeredVideoFileSuffixes
    "Answer the file extensions for which MPEGPlayer registers video services with FileList."
     "MPEGPlayer registeredVideoFileSuffixes"

     ^{ 'mpg'. 'mpeg'. 'jmv'. }

	! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:29'!
audioChannels: aStream
	^self external audioChannels: aStream! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 13:59'!
audioPlayerProcess
	^audioPlayerProcess ! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
audioPlayerProcess: aProcess
	audioPlayerProcess := aProcess! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:31'!
audioSampleRate: aStream
	^self external audioSampleRate: aStream! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:33'!
audioSamples: aStream
	^self external audioSamples: aStream
! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'JMM 11/8/2000 10:38'!
backAudio: aNumber forStream: aStream
	self forwardAudio: (0-aNumber) forStream: aStream! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'JMM 11/8/2000 10:39'!
backVideo: aNumber forStream: aStream
	self forwardVideo: (0-aNumber) forStream: aStream! !

!MPEGPlayer methodsFor: 'delay' stamp: 'stephaneducasse 2/4/2006 20:48'!
calculateDelayGivenFrame: frame stream: aStream
	| estimated current delta |

	current :=  Time millisecondClockValue  - (self startTimeForStream: aStream).
	estimated := ((frame asFloat / self frameRate) * 1000) asInteger  - (self clockBiasForStream: aStream).
	delta := estimated - current.
	delta > 33  ifTrue: 
		[self lastDelay: (delta + self lastDelay) // 2. 
		 (Delay forMilliseconds: self lastDelay) wait].
	delta < -33  ifTrue: 
		[self lastDelay: self lastDelay // 2.
		 self decideToSkipAFrame: delta averageWait: current//frame stream: aStream].
	! !

!MPEGPlayer methodsFor: 'delay' stamp: 'stephaneducasse 2/4/2006 20:48'!
calculateDelayToSoundGivenFrame: frame stream: aStream
	| current delta buffers estimatedAudio estimatedVideo |

	current :=  Time millisecondClockValue   - (self startTimeForStream: aStream) + (self clockBiasForStream: aStream).
	buffers := (self soundQueue sounds size - 1 ) max: 0.
	buffers = 0 ifTrue: [^self].
	estimatedAudio :=  ((self currentAudioSampleForStream: aStream) 
			- (buffers * self sampleRate // 10) 
			- self soundQueue currentSound samplesRemaining) * 1000 / self sampleRate.
	estimatedAudio := estimatedAudio - 0000.
	estimatedVideo := ((frame asFloat / self frameRate) * 1000) asInteger.
	delta := estimatedVideo - estimatedAudio.
	delta > 100  ifTrue: 
		[self lastDelay < delta ifTrue: [self lastDelay: self lastDelay + (((delta-self lastDelay)/10) max: 1)].
		(Delay forMilliseconds: self lastDelay) wait].
	delta < -100  ifTrue: 
		[self lastDelay: ((self lastDelay - 10) max: 1).
		 self decideToSkipAFrame: delta averageWait: current//frame stream: aStream].
! !

!MPEGPlayer methodsFor: 'utility' stamp: 'JMM 10/17/2000 23:22'!
changed
	self morph notNil ifTrue: [self morph changed].! !

!MPEGPlayer methodsFor: 'utility' stamp: 'stephaneducasse 2/4/2006 20:48'!
checkForm: aStream
	| y x |

	self form notNil ifTrue: [^self].
	y := self videoFrameHeight: aStream.
	x := self videoFrameWidth: aStream.
	self form:  (Form extent: x@y depth: 32)
! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 15:31'!
clockBias
	^clockBias! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
clockBias: aArray
	clockBias := aArray! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 15:31'!
clockBiasForStream: aStream
	^self clockBias at: aStream + 1.! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 15:32'!
clockBiasForStream: aStream put: aValue
	self clockBias at: aStream + 1 put: aValue! !

!MPEGPlayer methodsFor: 'initialize-release' stamp: 'JMM 11/8/2000 10:31'!
close
	self external closeFile! !

!MPEGPlayer methodsFor: 'audio' stamp: 'stephaneducasse 2/4/2006 20:48'!
createSoundFrom: aStream 
	| snds channels |
	
	snds := OrderedCollection new.
	channels := self audioChannels: 0.
	1 to: channels do: [:c | snds add: (self readSoundChannel: c - 1 stream: aStream)].
	channels = 1
		ifTrue:[^ MixedSound new
				add: (snds at: 1) pan: 0.5 volume: volume;
				
				yourself]
		ifFalse: [
			^ MixedSound new
				add: (snds at: 1) pan: 0.0 volume: volume;
				add: (snds at: 2) pan: 1.0 volume: volume;
				yourself].! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 11/8/2000 10:30'!
currentAudioSampleForStream: aStream
	^self external audioGetSample: aStream! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 11/8/2000 10:31'!
currentAudioSampleForStream: aStream put: aNumber
	self external audioSetSample: aNumber stream: aStream! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/13/2000 20:05'!
currentExternalLocationInPercent
	"Warning this might not return what you want, it gets percentage based on audio, or video stream based on last usage, because we buffer audio it may give incorrect information when playing mpeg movies"
	^self external getPercentage! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/13/2000 20:09'!
currentLocationInPercent: aStream

	self hasVideo ifTrue: [^ ((self currentVideoFrameForStream: aStream)/(self videoFrames: aStream)) asFloat].
	self hasAudio ifTrue: [^ ((self currentAudioSampleForStream: aStream)/(self audioSamples: aStream)) asFloat].

! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 11/8/2000 10:29'!
currentVideoFrameForStream: aStream
	^self external videoGetFrame: aStream! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 11/8/2000 10:30'!
currentVideoFrameForStream: aStream put: aNumber
	self external videoSetFrame: aNumber stream: aStream! !

!MPEGPlayer methodsFor: 'delay' stamp: 'stephaneducasse 2/4/2006 20:48'!
decideToSkipAFrame: delta averageWait: aWaitTime stream: aStream
	| estimatedFrames |

	delta abs > aWaitTime ifTrue: 
		[estimatedFrames := ( delta abs / (1000 / self frameRate)) asInteger.
		self videoDropFrames:  estimatedFrames stream: aStream].! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:32'!
endOfAudio: aStream
	^self external endOfAudio: aStream
! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:30'!
endOfVideo: aStream
	^self external endOfVideo: aStream
! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
errorForSoundStart: aValueInMilliseconds
	errorForSoundStart := aValueInMilliseconds
! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 1/20/2006 18:10'!
external
	[external hasVideo] on: Error do: 
		[self isBuffer
			ifTrue:
			[external := MPEGFile openBuffer: external buffer]
			ifFalse: 
				[(MPEGFile isFileValidMPEG: external fileName) 
					ifFalse: [^self error: 'Mpeg File is invalid'].
				external := MPEGFile openFile: external fileName]].
	^external! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 11/8/2000 10:31'!
fileName
	^self external fileName! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 10/17/2000 23:29'!
form
	form isNil ifTrue: 
		[self morph isNil ifTrue: [^nil].
		^self morph form].
	^form! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
form: aForm
	form := aForm! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'stephaneducasse 2/4/2006 20:48'!
forwardAudio: aNumber forStream: aStream
	| newLocation |

	self hasAudio ifFalse: [^self].
	newLocation := (((self currentAudioSampleForStream: aStream) + aNumber) min: (self audioSamples: aStream)) max: 0 .
	self currentAudioSampleForStream: aStream put: newLocation! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'stephaneducasse 2/4/2006 20:48'!
forwardVideo: aNumber forStream: aStream
	| newLocation |

	self hasVideo ifFalse: [^self].
	newLocation := (((self currentVideoFrameForStream: aStream) + aNumber) min: (self videoFrames: aStream)) max: 0.
	self currentVideoFrameForStream: aStream put: newLocation.
! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/19/2000 17:39'!
frameRate
	^frameRate! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
frameRate: aRate
	frameRate := aRate! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:33'!
getTOC: timecode doStreams: streams
	^self external getTOC: timecode doStreams: streams
! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:31'!
getTimeCode
	^self external getTimeCode! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:31'!
hasAudio
	^self external hasAudio
	! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:30'!
hasVideo
	^self external hasVideo! !

!MPEGPlayer methodsFor: 'initialize-release' stamp: 'muOuser 4/18/2007 10:48'!
initialize: aPath 
	
	self isBuffer
		ifTrue: [external := MPEGFile openBuffer: buffer]
		ifFalse: [(MPEGFile isFileValidMPEG: aPath)
				ifFalse: [^ nil].
			external := MPEGFile openFile: aPath].
	self playerProcessPriority: Processor userSchedulingPriority.
	self lastDelay: 10.
	volume := 1.0.
	errorForSoundStart := 500.
	semaphoreForSound := Semaphore new.
	self
		startTime: (Array new: self totalVideoStreams).
	self
		clockBias: (Array new: self totalVideoStreams withAll: 0)! !

!MPEGPlayer methodsFor: 'initialize-release' stamp: 'JMM 1/20/2006 18:17'!
initializeWithBuffer: aBuffer form: aForm
	isBuffer := true.
	buffer := aBuffer.
	self initialize: aBuffer.
	self form: aForm.
	^self! !

!MPEGPlayer methodsFor: 'initialize-release' stamp: 'JMM 1/20/2006 18:07'!
initializeWithBuffer: aBuffer morph: aMorphic
	isBuffer := true.
	buffer := aBuffer.
	self initialize: aBuffer.
	self morph: aMorphic.
	^self! !

!MPEGPlayer methodsFor: 'initialize-release' stamp: 'JMM 9/20/2000 14:06'!
initializeWithFileName: aPath 
	self initialize: aPath.
	self form: nil.
	^self! !

!MPEGPlayer methodsFor: 'initialize-release' stamp: 'JMM 9/20/2000 14:06'!
initializeWithFileName: aPath form: aForm
	self initialize: aPath.
	self form: aForm.
	^self! !

!MPEGPlayer methodsFor: 'initialize-release' stamp: 'JMM 10/17/2000 23:34'!
initializeWithFileName: aPath morph: aMorphic
	self initialize: aPath.
	self morph: aMorphic.
	^self! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 1/20/2006 18:05'!
isBuffer
	^isBuffer == true! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'JMM 11/19/2000 12:50'!
isPlaying
	^((self audioPlayerProcess isNil) and: [self videoPlayerProcess isNil]) not! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 14:04'!
lastDelay
	^lastDelay! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
lastDelay: aNumber
	lastDelay := aNumber! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 10/17/2000 23:20'!
morph
	^morph! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
morph: aMorph	
	morph := aMorph.! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 10/20/2000 22:36'!
mpegFile
	^external! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 13:59'!
noSound
	^noSound! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
noSound: flag
	noSound := flag! !

!MPEGPlayer methodsFor: 'play' stamp: 'JMM 11/8/2000 10:20'!
playAudioStream: aStream

	self hasAudio ifFalse: [^self].
	self setupStream: aStream.
	self startAudioPlayerProcess: aStream.! !

!MPEGPlayer methodsFor: 'play' stamp: 'JMM 11/8/2000 10:20'!
playAudioStreamNoSeek: aStream

	self hasAudio ifFalse: [^self].
	self setupStreamNoSeek: aStream.
	self startAudioPlayerProcess: aStream.! !

!MPEGPlayer methodsFor: 'play' stamp: 'JMM 11/8/2000 10:20'!
playAudioStreamWaitTilDone: aStream

	self hasAudio ifFalse: [^self].
	self setupStream: aStream.
	self privatePlayAudioStream: aStream.! !

!MPEGPlayer methodsFor: 'play' stamp: 'JMM 11/8/2000 10:20'!
playStream: aStream
	self noSound: self hasAudio not.
	self startVideoPlayerProcess: aStream
! !

!MPEGPlayer methodsFor: 'play' stamp: 'JMM 11/8/2000 10:20'!
playStreamWaitTilDone: aStream
	self noSound: self hasAudio not.
	self privatePlayVideoStream: aStream.! !

!MPEGPlayer methodsFor: 'play' stamp: 'JMM 9/20/2000 14:00'!
playVideoStream: aStream
	self noSound: true.
	self startVideoPlayerProcess: aStream

! !

!MPEGPlayer methodsFor: 'play' stamp: 'JMM 9/20/2000 14:00'!
playVideoStreamWaitTilDone: aStream
	self noSound: true.
	self privatePlayVideoStream: aStream

! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/19/2000 17:59'!
playerProcessPriority
	^playerProcessPriority! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
playerProcessPriority: aNumber
	playerProcessPriority := aNumber! !

!MPEGPlayer methodsFor: 'audio' stamp: 'stephaneducasse 2/4/2006 20:48'!
privatePlayAudioStream: aStream
	| number |

	number := 5.
	self soundQueue: (QueueSound new startTime: 0).
	[number + 2 timesRepeat: [self soundQueue add: (self createSoundFrom: aStream)].
	self soundQueue play.
	semaphoreForSound signal.
	[[self soundQueue sounds size > number] whileTrue: [(Delay forMilliseconds: 100) wait].
	self soundQueue add: (self createSoundFrom: aStream).
	(self endOfAudio: aStream) 
		ifTrue: 
			[self audioPlayerProcess: nil.
			^self]] repeat] on: Error do: 
				[self audioPlayerProcess: nil.
				^self]! !

!MPEGPlayer methodsFor: 'video' stamp: 'stephaneducasse 2/4/2006 20:48'!
privatePlayVideoStream: aStream
	
	| location |
	self hasVideo ifFalse: 
		[self timeCheck: 0@0.
		^self].
	self checkForm: aStream.
	self frameRate: (self videoFrameRate: aStream).
	location := self currentVideoFrameForStream: aStream.
	self clockBiasForStream: aStream 
		put: (1/self frameRate*location*1000) asInteger.
	self videoLoop: aStream.
	self timeCheck: ((Time millisecondClockValue + (self clockBiasForStream: aStream) - (self startTimeForStream: aStream))/1000.0) @ ((self videoFrames: aStream) / self frameRate).
	self videoPlayerProcess: nil! !

!MPEGPlayer methodsFor: 'audio' stamp: 'fbs 2/13/2006 22:44'!
readSoundChannel: aChannel stream: aStream
	| buffer result samples |

	samples := (self sampleRate // 10)  min: 
		((self audioSamples: aStream) - (self currentAudioSampleForStream: aStream)).
	(samples = 0) ifTrue: [self error: 'Mpeg at end of stream, toss error, catch up high']. 
	buffer := SoundBuffer newMonoSampleCount: samples.
	aChannel = 0 
		ifTrue: [result := self external audioReadBuffer: buffer stream: 
					aStream channel: aChannel]
		ifFalse: [result := self external audioReReadBuffer: buffer stream: 
					aStream channel: aChannel].
	^SampledSound samples: buffer samplingRate: self sampleRate.
! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'stephaneducasse 2/4/2006 20:48'!
recalculateNewSampleLocationForStream: aStream givenFrame: aFrame
	| estimated |

	self hasAudio ifFalse: [^self].
	estimated := (aFrame / (self videoFrames: aStream) * (self audioSampleRate: aStream)) asInteger.
	self currentAudioSampleForStream: aStream put: estimated.! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/19/2000 17:34'!
sampleRate
	^sampleRate! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
sampleRate: aRate
	sampleRate := aRate! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'JMM 11/8/2000 10:47'!
seekToHere: aPercentage forStream: aStream
	"Alternate method is to seek all video/audio for stream to a certain percentage using the primitive, but I think your mpeg must have timecodes!! otherwise endless loop"
	self external seekPercentage: aPercentage! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'JMM 11/19/2000 11:44'!
seekVideoAudioBasedOnFrame: aFrame forStream: aStream
	self external hasVideo ifTrue: 
		[self currentVideoFrameForStream: aStream put:  aFrame].
	self recalculateNewSampleLocationForStream: aStream givenFrame: aFrame! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'JMM 11/10/2000 00:19'!
setLocation: aPercentage forStream: aStream
	self hasAudio ifTrue: [self currentAudioSampleForStream: aStream put: ((self audioSamples: aStream) * aPercentage) asInteger]. 
	self hasVideo ifTrue: [self currentVideoFrameForStream: aStream put: ((self videoFrames: aStream) * aPercentage) asInteger].! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:31'!
setMMX: aBoolean
	self external setMMX: aBoolean! !

!MPEGPlayer methodsFor: 'audio' stamp: 'JMM 11/8/2000 10:49'!
setupStream: aStream
	self sampleRate: (self audioSampleRate: aStream).
	SoundPlayer startPlayerProcessBufferSize:  8192 "(SoundPlayer bufferMSecs * self sampleRate) // 1000"
		rate: self sampleRate stereo: true.
! !

!MPEGPlayer methodsFor: 'audio' stamp: 'JMM 11/8/2000 10:33'!
setupStreamNoSeek: aStream
	self sampleRate: (self audioSampleRate: aStream).
	SoundPlayer startPlayerProcessBufferSize:  8192 "(SoundPlayer bufferMSecs * self sampleRate) // 1000"
		rate: self sampleRate stereo: ((self  audioChannels: aStream) > 1).
! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 14:01'!
soundQueue
	^soundQueue! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
soundQueue: aQueue
	soundQueue := aQueue! !

!MPEGPlayer methodsFor: 'audio' stamp: 'JMM 9/20/2000 13:38'!
startAudioPlayerProcess: aStream
	self audioPlayerProcess: ([self privatePlayAudioStream: aStream] forkAt: Processor userInterruptPriority)! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 14:38'!
startTime
	^startTime! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
startTime: aArray
	startTime := aArray! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 15:12'!
startTimeForStream: aStream
	^self startTime at: aStream + 1! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 15:11'!
startTimeForStream: aStream put: aNumber
	^self startTime at: aStream + 1 put: aNumber! !

!MPEGPlayer methodsFor: 'video' stamp: 'JMM 9/20/2000 13:59'!
startVideoPlayerProcess: aStream
	self videoPlayerProcess: ([self privatePlayVideoStream: aStream] forkAt: self playerProcessPriority)! !

!MPEGPlayer methodsFor: 'play controls' stamp: 'JMM 9/20/2000 18:46'!
stop
	self videoPlayerProcess notNil ifTrue: 
		[self videoPlayerProcess terminate. 
		self videoPlayerProcess: nil].
	self audioPlayerProcess notNil ifTrue: 
		[self audioPlayerProcess terminate. 
		self audioPlayerProcess: nil.
		SoundPlayer stopPlayingAll]! !

!MPEGPlayer methodsFor: 'initialize-release' stamp: 'JMM 9/20/2000 14:15'!
stopAndClose
	self stop.
	self close
! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 14:03'!
timeCheck
	^timeCheck! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
timeCheck: aNumber
	timeCheck := aNumber! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'bh 12/11/2000 22:17'!
totalSeconds

	^(self audioSamples: 0) / (self audioSampleRate: 0)! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:33'!
totalVideoStreams
	^self external totalVideoStreams
! !

!MPEGPlayer methodsFor: 'audio' stamp: 'JMM 9/19/2000 16:59'!
updateSoundStream: aStream! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:32'!
videoDropFrames: aNumberOfFrames stream: aStream
	^self external videoDropFrames: aNumberOfFrames stream: aStream! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:30'!
videoFrameHeight: aStream
	^self external videoFrameHeight: aStream
! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:29'!
videoFrameRate: aStream
	^self external videoFrameRate: aStream! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:29'!
videoFrameWidth: aStream
	^self external videoFrameWidth: aStream
! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:33'!
videoFrames: aStream
	^self external videoFrames: aStream
! !

!MPEGPlayer methodsFor: 'video' stamp: 'stephaneducasse 2/4/2006 20:48'!
videoLoop: aStream
	| location oneTime | 

	oneTime := true.
	[self external videoReadFrameInto: self form stream: aStream.
	oneTime ifTrue: 
			[oneTime := false.
			self noSound ifFalse: 
				[self playAudioStreamNoSeek: aStream.
				semaphoreForSound wait.
				(Delay forMilliseconds: errorForSoundStart) wait].
			self startTimeForStream: aStream put: (Time millisecondClockValue)].
	self morph ifNil: 
			[self form == Display
				ifTrue: [Display forceToScreen]
				ifFalse: [self form displayOn: Display]].
	self changed.
		location := (self currentVideoFrameForStream: aStream)+1.
	true 
			ifTrue: [self calculateDelayGivenFrame: location stream: aStream]
			ifFalse: [self calculateDelayToSoundGivenFrame: location stream: aStream].
	(self endOfVideo: aStream)  ifTrue: [^self]] repeat.! !

!MPEGPlayer methodsFor: 'access' stamp: 'JMM 9/20/2000 13:58'!
videoPlayerProcess
	^videoPlayerProcess ! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
videoPlayerProcess: aProcess
	videoPlayerProcess := aProcess! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:33'!
videoPreviousFrame: aStream
	^self external videoPreviousFrame: aStream
! !

!MPEGPlayer methodsFor: 'access to attributes' stamp: 'JMM 11/8/2000 10:33'!
videoSetCPUs: aNumber
	^self external videoSetCPUs: aNumber
! !

!MPEGPlayer methodsFor: 'access' stamp: 'stephaneducasse 2/4/2006 20:48'!
volume: aVolume
	volume := aVolume! !

Object subclass: #MPEGSubtitleElement
	instanceVariableNames: 'initialFrame endFrame subtitleLine contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Kernel'!

!MPEGSubtitleElement commentStamp: 'asm 7/31/2003 22:27' prior: 0!
an element of a subtitle file,

this has the form

{initialFrame}{endFrame} subtitle line[| next subtitle line]!

!MPEGSubtitleElement class methodsFor: 'instance creation' stamp: 'asm 7/30/2003 21:26'!
fromStream: aStream
	"Returns an instance of the receiver read from aStream."

	^self new readFrom: aStream! !

!MPEGSubtitleElement methodsFor: 'accessing' stamp: 'dgd 3/8/2004 20:17'!
contents
	"answer the receiver's contents"
	^ contents! !

!MPEGSubtitleElement methodsFor: 'accessing' stamp: 'dgd 3/8/2004 20:17'!
contents: aString 
	"change the receiver's contents"
	contents := aString replaceAll: $| with: Character cr! !

!MPEGSubtitleElement methodsFor: 'testing' stamp: 'dgd 3/8/2004 20:23'!
correspondsToFrame: aNumber
	"answer if the receiver corresponds to a given frame number"
	^ aNumber between:  initialFrame and:  endFrame! !

!MPEGSubtitleElement methodsFor: 'parsing' stamp: 'asm 7/30/2003 21:04'!
is: text in: aStream
	" Returns true if text is present in aStream.
	Advance the stream if present. "

	| position |
	(text isKindOf: Character) ifTrue: [
		^self is: (String with: text) in: aStream
	].
	position := aStream position.
	aStream skipSeparators.
	text = (aStream next: text size) ifFalse: [
		aStream position: position.
		^false
	].
	^true! !

!MPEGSubtitleElement methodsFor: 'parsing' stamp: 'asm 7/30/2003 21:01'!
mustBe: text in: aStream
	" Check text to be present in aStream. "

	(text isKindOf: Character) ifTrue: [
		^self is: (String with: text) in: aStream
	].
	(self is: text in: aStream) ifFalse: [
		^self error: 'Invalid token, must be: ',text
	].! !

!MPEGSubtitleElement methodsFor: 'parsing' stamp: 'asm 7/30/2003 21:05'!
nextIntegerFrom: aStream
	" Returns the next Integer present in aStream. "

	| sign result |
	sign := (self is: $- in: aStream) ifTrue: [-1] ifFalse: [1].
	result := 0.
	self skipBlanks: aStream.
	[aStream peek isDigit] whileTrue: [
		result := aStream next asciiValue - $0 asciiValue + (result * 10)
	].
	^result * sign! !

!MPEGSubtitleElement methodsFor: 'printing' stamp: 'dgd 3/8/2004 20:50'!
printOn: aStream 
	"append to aStream a sequence of characters that identifies 
	the receiver."
	aStream nextPutAll: '{';
		 nextPutAll: initialFrame asString;
		 nextPutAll: '}{';
		 nextPutAll: endFrame asString;
		 nextPutAll: '}';
		 nextPutAll: contents asString! !

!MPEGSubtitleElement methodsFor: 'parsing' stamp: 'dgd 3/8/2004 20:17'!
readFrom: aStream 
	"Private - Read the receiver's contents from aStream."
	self mustBe: '{' in: aStream.
	initialFrame := self nextIntegerFrom: aStream.
	self mustBe: '}{' in: aStream.
	endFrame := self nextIntegerFrom: aStream.
	self mustBe: '}' in: aStream.
	""
	self contents: aStream nextLine isoToSqueak! !

!MPEGSubtitleElement methodsFor: 'parsing' stamp: 'asm 7/30/2003 21:42'!
skipBlanks: aStream
	" Advance aStream skipping all blank characters and comments. "

	aStream skipSeparators! !

Object subclass: #MPEGSubtitles
	instanceVariableNames: 'fileName elements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Movies-Kernel'!

!MPEGSubtitles commentStamp: 'asm 7/31/2003 22:12' prior: 0!
a subtitle file

i can only read subtitle files with a format like this:

[..]
{1043}{1082}La gente siempre me pregunta|si conozco a Tyler Durden.
{1083}{1096}Tres minutos.
{1097}{1133}El momento de la verdad.|Punto cero.
[..]
from Fight Club

while reading, pipes(|) are replaced by carriage returns
!

!MPEGSubtitles class methodsFor: 'instance creation' stamp: 'dgd 3/8/2004 22:02'!
fromFileNamed: aString 
	"Returns an instance of the receiver read from file named  
	aString"
	^self new initializeFromFileNamed: aString ! !

!MPEGSubtitles methodsFor: 'accessing' stamp: 'dgd 3/8/2004 20:49'!
elementCorrespondingToFrame: frameNumber 
	"answer the element corresponding to frameNumber"
	^ elements
		detect: [:each | each correspondsToFrame: frameNumber]
		ifNone: []! !

!MPEGSubtitles methodsFor: 'accessing' stamp: 'dgd 3/8/2004 22:45'!
fileName
	"answer the receiver's fileName"
	^ fileName! !

!MPEGSubtitles methodsFor: 'initialization' stamp: 'dgd 3/8/2004 22:24'!
initializeFromFileNamed: aString 
	"initialize the receiver from a file named aString"
	| file result |
fileName := aString.
	elements := OrderedCollection new.
	""
	file := CrLfFileStream readOnlyFileNamed: aString.
	[result := self readFrom: file]
		ensure: [file close].
	^ result! !

!MPEGSubtitles methodsFor: 'initialization' stamp: 'dgd 3/8/2004 22:04'!
readFrom: aStream 
	"private - Read the next definitions found in aStream onto the  
	receiver"
	[aStream atEnd]
		whileFalse: [| element | 
			element := MPEGSubtitleElement fromStream: aStream.
			elements add: element]! !

!MPEGSubtitles methodsFor: 'accessing' stamp: 'dgd 3/8/2004 20:42'!
subtitleForFrame: frameNumber 
	"answer the subtitle for the given frame number"
	| element |
	element := self elementCorrespondingToFrame: frameNumber.
	^ element isNil
		ifTrue: ['']
		ifFalse: [element contents]! !

!PianoRollScoreMorph methodsFor: '*movies' stamp: 'di 10/4/2000 16:40'!
addMovieClipPlayer

	movieClipPlayer _ MoviePlayerMorph new.
	movieClipPlayer pianoRoll: self.  "back link"
	self activeHand attachMorph: movieClipPlayer! !
MoviePlayerMorph initialize!
MPEGMoviePlayerMorph initialize!
